#https://github.com/scikit-learn/scikit-learn/blob/14031f6/sklearn/preprocessing/data.py#L1139
# Authors: Alexandre Gramfort <alexandre.gramfort@inria.fr>
#          Mathieu Blondel <mathieu@mblondel.org>
#          Olivier Grisel <olivier.grisel@ensta.org>
#          Andreas Mueller <amueller@ais.uni-bonn.de>
#          Eric Martin <eric@ericmart.in>
#          Giorgio Patrini <giorgio.patrini@anu.edu.au>
# License: BSD 3 clause

from itertools import chain, combinations
import numbers
import warnings

import numpy as np
from scipy import sparse
import copy as PythonCopy

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.externals import six
from sklearn.utils import check_array
from sklearn.utils import deprecated
from sklearn.utils.extmath import row_norms
from sklearn.utils.extmath import _incremental_mean_and_var
from itertools import combinations_with_replacement as combinations_w_r
#from sklearn.utils.fixes import bincount
from numpy import bincount
from sklearn.utils.sparsefuncs_fast import (inplace_csr_row_normalize_l1,
                                      inplace_csr_row_normalize_l2)
from sklearn.utils.sparsefuncs import (inplace_column_scale,
                                 mean_variance_axis, incr_mean_variance_axis,
                                 min_max_axis)
from sklearn.utils.validation import check_is_fitted, FLOAT_DTYPES


zip = six.moves.zip
map = six.moves.map
range = six.moves.range

__all__ = [
    'Binarizer',
    'KernelCenterer',
    'MinMaxScaler',
    'MaxAbsScaler',
    'Normalizer',
    'OneHotEncoder',
    'RobustScaler',
    'StandardScaler',
    'add_dummy_feature',
    'binarize',
    'normalize',
    'scale',
    'robust_scale',
    'maxabs_scale',
    'minmax_scale',
]

DEPRECATION_MSG_1D = (
    "Passing 1d arrays as data is deprecated in 0.17 and will "
    "raise ValueError in 0.19. Reshape your data either using "
    "X.reshape(-1, 1) if your data has a single feature or "
    "X.reshape(1, -1) if it contains a single sample."
)


def _handle_zeros_in_scale(scale, copy=True):
    ''' Makes sure that whenever scale is zero, we handle it correctly.
    This happens in most scalers when we have constant features.'''

    # if we are fitting on 1D arrays, scale might be a scalar
    if np.isscalar(scale):
        if scale == .0:
            scale = 1.
        return scale
    elif isinstance(scale, np.ndarray):
        if copy:
            # New array to avoid side-effects
            scale = scale.copy()
        scale[scale == 0.0] = 1.0
        return scale


def scale(X, axis=0, with_mean=True, with_std=True, copy=True):
    """Standardize a dataset along any axis
    Center to the mean and component wise scale to unit variance.
    Read more in the :ref:`User Guide <preprocessing_scaler>`.
    Parameters
    ----------
    X : {array-like, sparse matrix}
        The data to center and scale.
    axis : int (0 by default)
        axis used to compute the means and standard deviations along. If 0,
        independently standardize each feature, otherwise (if 1) standardize
        each sample.
    with_mean : boolean, True by default
        If True, center the data before scaling.
    with_std : boolean, True by default
        If True, scale the data to unit variance (or equivalently,
        unit standard deviation).
    copy : boolean, optional, default True
        set to False to perform inplace row normalization and avoid a
        copy (if the input is already a numpy array or a scipy.sparse
        CSC matrix and if axis is 1).
    Notes
    -----
    This implementation will refuse to center scipy.sparse matrices
    since it would make them non-sparse and would potentially crash the
    program with memory exhaustion problems.
    Instead the caller is expected to either set explicitly
    `with_mean=False` (in that case, only variance scaling will be
    performed on the features of the CSC matrix) or to call `X.toarray()`
    if he/she expects the materialized dense array to fit in memory.
    To avoid memory copy the caller should pass a CSC matrix.
    See also
    --------
    StandardScaler: Performs scaling to unit variance using the``Transformer`` API
        (e.g. as part of a preprocessing :class:`sklearn.pipeline.Pipeline`).
    """  # noqa
    X = check_array(X, accept_sparse='csc', copy=copy, ensure_2d=False,
                    warn_on_dtype=True, estimator='the scale function',
                    dtype=FLOAT_DTYPES)
    if sparse.issparse(X):
        if with_mean:
            raise ValueError(
                "Cannot center sparse matrices: pass `with_mean=False` instead"
                " See docstring for motivation and alternatives.")
        if axis != 0:
            raise ValueError("Can only scale sparse matrix on axis=0, "
                             " got axis=%d" % axis)
        if with_std:
            _, var = mean_variance_axis(X, axis=0)
            var = _handle_zeros_in_scale(var, copy=False)
            inplace_column_scale(X, 1 / np.sqrt(var))
    else:
        X = np.asarray(X)
        if with_mean:
            mean_ = np.mean(X, axis)
        if with_std:
            scale_ = np.std(X, axis)
        # Xr is a view on the original array that enables easy use of
        # broadcasting on the axis in which we are interested in
        Xr = np.rollaxis(X, axis)
        if with_mean:
            Xr -= mean_
            mean_1 = Xr.mean(axis=0)
            # Verify that mean_1 is 'close to zero'. If X contains very
            # large values, mean_1 can also be very large, due to a lack of
            # precision of mean_. In this case, a pre-scaling of the
            # concerned feature is efficient, for instance by its mean or
            # maximum.
            if not np.allclose(mean_1, 0):
                warnings.warn("Numerical issues were encountered "
                              "when centering the data "
                              "and might not be solved. Dataset may "
                              "contain too large values. You may need "
                              "to prescale your features.")
                Xr -= mean_1
        if with_std:
            scale_ = _handle_zeros_in_scale(scale_, copy=False)
            Xr /= scale_
            if with_mean:
                mean_2 = Xr.mean(axis=0)
                # If mean_2 is not 'close to zero', it comes from the fact that
                # scale_ is very small so that mean_2 = mean_1/scale_ > 0, even
                # if mean_1 was close to zero. The problem is thus essentially
                # due to the lack of precision of mean_. A solution is then to
                # subtract the mean again:
                if not np.allclose(mean_2, 0):
                    warnings.warn("Numerical issues were encountered "
                                  "when scaling the data "
                                  "and might not be solved. The standard "
                                  "deviation of the data is probably "
                                  "very close to 0. ")
                    Xr -= mean_2
    return X


class MinMaxScaler(BaseEstimator, TransformerMixin):
    """Transforms features by scaling each feature to a given range.
    This estimator scales and translates each feature individually such
    that it is in the given range on the training set, i.e. between
    zero and one.
    The transformation is given by::
        X_std = (X - X.min(axis=0)) / (X.max(axis=0) - X.min(axis=0))
        X_scaled = X_std * (max - min) + min
    where min, max = feature_range.
    This transformation is often used as an alternative to zero mean,
    unit variance scaling.
    Read more in the :ref:`User Guide <preprocessing_scaler>`.
    Parameters
    ----------
    feature_range : tuple (min, max), default=(0, 1)
        Desired range of transformed data.
    copy : boolean, optional, default True
        Set to False to perform inplace row normalization and avoid a
        copy (if the input is already a numpy array).
    Attributes
    ----------
    min_ : ndarray, shape (n_features,)
        Per feature adjustment for minimum.
    scale_ : ndarray, shape (n_features,)
        Per feature relative scaling of the data.
        .. versionadded:: 0.17
           *scale_* attribute.
    data_min_ : ndarray, shape (n_features,)
        Per feature minimum seen in the data
        .. versionadded:: 0.17
           *data_min_* instead of deprecated *data_min*.
    data_max_ : ndarray, shape (n_features,)
        Per feature maximum seen in the data
        .. versionadded:: 0.17
           *data_max_* instead of deprecated *data_max*.
    data_range_ : ndarray, shape (n_features,)
        Per feature range ``(data_max_ - data_min_)`` seen in the data
        .. versionadded:: 0.17
           *data_range_* instead of deprecated *data_range*.
    See also
    --------
    minmax_scale: Equivalent function without the object oriented API.
    """

    def __init__(self, feature_range=(0, 1), copy=True):
        self.feature_range = feature_range
        self.copy = copy

    @property
    @deprecated("Attribute data_range will be removed in "
                "0.19. Use ``data_range_`` instead")
    def data_range(self):
        return self.data_range_

    @property
    @deprecated("Attribute data_min will be removed in "
                "0.19. Use ``data_min_`` instead")
    def data_min(self):
        return self.data_min_

    def _reset(self):
        """Reset internal data-dependent state of the scaler, if necessary.
        __init__ parameters are not touched.
        """

        # Checking one attribute is enough, becase they are all set together
        # in partial_fit
        if hasattr(self, 'scale_'):
            del self.scale_
            del self.min_
            del self.n_samples_seen_
            del self.data_min_
            del self.data_max_
            del self.data_range_

    def fit(self, X, y=None):
        """Compute the minimum and maximum to be used for later scaling.
        Parameters
        ----------
        X : array-like, shape [n_samples, n_features]
            The data used to compute the per-feature minimum and maximum
            used for later scaling along the features axis.
        """

        # Reset internal state before fitting
        self._reset()
        return self.partial_fit(X, y)

    def partial_fit(self, X, y=None):
        """Online computation of min and max on X for later scaling.
        All of X is processed as a single batch. This is intended for cases
        when `fit` is not feasible due to very large number of `n_samples`
        or because X is read from a continuous stream.
        Parameters
        ----------
        X : array-like, shape [n_samples, n_features]
            The data used to compute the mean and standard deviation
            used for later scaling along the features axis.
        y : Passthrough for ``Pipeline`` compatibility.
        """
        feature_range = self.feature_range
        if feature_range[0] >= feature_range[1]:
            raise ValueError("Minimum of desired feature range must be smaller"
                             " than maximum. Got %s." % str(feature_range))

        if sparse.issparse(X):
            raise TypeError("MinMaxScaler does no support sparse input. "
                            "You may consider to use MaxAbsScaler instead.")

        X = check_array(X, copy=self.copy, ensure_2d=False, warn_on_dtype=True,
                        estimator=self, dtype=FLOAT_DTYPES)

        if X.ndim == 1:
            warnings.warn(DEPRECATION_MSG_1D, DeprecationWarning)

        data_min = np.min(X, axis=0)
        data_max = np.max(X, axis=0)

        # First pass
        if not hasattr(self, 'n_samples_seen_'):
            self.n_samples_seen_ = X.shape[0]
        # Next steps
        else:
            data_min = np.minimum(self.data_min_, data_min)
            data_max = np.maximum(self.data_max_, data_max)
            self.n_samples_seen_ += X.shape[0]

        data_range = data_max - data_min
        self.scale_ = ((feature_range[1] - feature_range[0]) /
                       _handle_zeros_in_scale(data_range))
        self.min_ = feature_range[0] - data_min * self.scale_
        self.data_min_ = data_min
        self.data_max_ = data_max
        self.data_range_ = data_range
        return self

    def transform(self, X):
        """Scaling features of X according to feature_range.
        Parameters
        ----------
        X : array-like, shape [n_samples, n_features]
            Input data that will be transformed.
        """
        check_is_fitted(self, 'scale_')

        X = check_array(X, copy=self.copy, ensure_2d=False, dtype=FLOAT_DTYPES)
        if X.ndim == 1:
            warnings.warn(DEPRECATION_MSG_1D, DeprecationWarning)

        X *= self.scale_
        X += self.min_
        return X

    def inverse_transform(self, X):
        """Undo the scaling of X according to feature_range.
        Parameters
        ----------
        X : array-like, shape [n_samples, n_features]
            Input data that will be transformed. It cannot be sparse.
        """
        check_is_fitted(self, 'scale_')

        X = check_array(X, copy=self.copy, ensure_2d=False, dtype=FLOAT_DTYPES)
        if X.ndim == 1:
            warnings.warn(DEPRECATION_MSG_1D, DeprecationWarning)

        X -= self.min_
        X /= self.scale_
        return X


def minmax_scale(X, feature_range=(0, 1), axis=0, copy=True):
    """Transforms features by scaling each feature to a given range.
    This estimator scales and translates each feature individually such
    that it is in the given range on the training set, i.e. between
    zero and one.
    The transformation is given by::
        X_std = (X - X.min(axis=0)) / (X.max(axis=0) - X.min(axis=0))
        X_scaled = X_std * (max - min) + min
    where min, max = feature_range.
    This transformation is often used as an alternative to zero mean,
    unit variance scaling.
    Read more in the :ref:`User Guide <preprocessing_scaler>`.
    .. versionadded:: 0.17
       *minmax_scale* function interface
       to :class:`sklearn.preprocessing.MinMaxScaler`.
    Parameters
    ----------
    feature_range : tuple (min, max), default=(0, 1)
        Desired range of transformed data.
    axis : int (0 by default)
        axis used to scale along. If 0, independently scale each feature,
        otherwise (if 1) scale each sample.
    copy : boolean, optional, default is True
        Set to False to perform inplace scaling and avoid a copy (if the input
        is already a numpy array).
    See also
    --------
    MinMaxScaler: Performs scaling to a given range using the``Transformer`` API
        (e.g. as part of a preprocessing :class:`sklearn.pipeline.Pipeline`).
    """  # noqa
    # To allow retro-compatibility, we handle here the case of 1D-input
    # From 0.17, 1D-input are deprecated in scaler objects
    # Although, we want to allow the users to keep calling this function
    # with 1D-input.

    # Cast input to array, as we need to check ndim. Prior to 0.17, that was
    # done inside the scaler object fit_transform.
    # If copy is required, it will be done inside the scaler object.
    X = check_array(X, copy=False, ensure_2d=False, warn_on_dtype=True,
                    dtype=FLOAT_DTYPES)
    original_ndim = X.ndim

    if original_ndim == 1:
        X = X.reshape(X.shape[0], 1)

    s = MinMaxScaler(feature_range=feature_range, copy=copy)
    if axis == 0:
        X = s.fit_transform(X)
    else:
        X = s.fit_transform(X.T).T

    if original_ndim == 1:
        X = X.ravel()

    return X


class StandardScaler(BaseEstimator, TransformerMixin):
    """Standardize features by removing the mean and scaling to unit variance
    Centering and scaling happen independently on each feature by computing
    the relevant statistics on the samples in the training set. Mean and
    standard deviation are then stored to be used on later data using the
    `transform` method.
    Standardization of a dataset is a common requirement for many
    machine learning estimators: they might behave badly if the
    individual feature do not more or less look like standard normally
    distributed data (e.g. Gaussian with 0 mean and unit variance).
    For instance many elements used in the objective function of
    a learning algorithm (such as the RBF kernel of Support Vector
    Machines or the L1 and L2 regularizers of linear models) assume that
    all features are centered around 0 and have variance in the same
    order. If a feature has a variance that is orders of magnitude larger
    that others, it might dominate the objective function and make the
    estimator unable to learn from other features correctly as expected.
    This scaler can also be applied to sparse CSR or CSC matrices by passing
    `with_mean=False` to avoid breaking the sparsity structure of the data.
    Read more in the :ref:`User Guide <preprocessing_scaler>`.
    Parameters
    ----------
    with_mean : boolean, True by default
        If True, center the data before scaling.
        This does not work (and will raise an exception) when attempted on
        sparse matrices, because centering them entails building a dense
        matrix which in common use cases is likely to be too large to fit in
        memory.
    with_std : boolean, True by default
        If True, scale the data to unit variance (or equivalently,
        unit standard deviation).
    copy : boolean, optional, default True
        If False, try to avoid a copy and do inplace scaling instead.
        This is not guaranteed to always work inplace; e.g. if the data is
        not a NumPy array or scipy.sparse CSR matrix, a copy may still be
        returned.
    Attributes
    ----------
    scale_ : ndarray, shape (n_features,)
        Per feature relative scaling of the data.
        .. versionadded:: 0.17
           *scale_* is recommended instead of deprecated *std_*.
    mean_ : array of floats with shape [n_features]
        The mean value for each feature in the training set.
    var_ : array of floats with shape [n_features]
        The variance for each feature in the training set. Used to compute
        `scale_`
    n_samples_seen_ : int
        The number of samples processed by the estimator. Will be reset on
        new calls to fit, but increments across ``partial_fit`` calls.
    See also
    --------
    scale: Equivalent function without the object oriented API.
    :class:`sklearn.decomposition.PCA`
        Further removes the linear correlation across features with 'whiten=True'.
    """  # noqa

    def __init__(self, copy=True, with_mean=True, with_std=True):
        self.with_mean = with_mean
        self.with_std = with_std
        self.copy = copy

    @property
    @deprecated("Attribute ``std_`` will be removed in 0.19. "
                "Use ``scale_`` instead")
    def std_(self):
        return self.scale_

    def _reset(self):
        """Reset internal data-dependent state of the scaler, if necessary.
        __init__ parameters are not touched.
        """

        # Checking one attribute is enough, becase they are all set together
        # in partial_fit
        if hasattr(self, 'scale_'):
            del self.scale_
            del self.n_samples_seen_
            del self.mean_
            del self.var_

    def fit(self, X, y=None):
        """Compute the mean and std to be used for later scaling.
        Parameters
        ----------
        X : {array-like, sparse matrix}, shape [n_samples, n_features]
            The data used to compute the mean and standard deviation
            used for later scaling along the features axis.
        y: Passthrough for ``Pipeline`` compatibility.
        """

        # Reset internal state before fitting
        self._reset()
        return self.partial_fit(X, y)

    def partial_fit(self, X, y=None):
        """Online computation of mean and std on X for later scaling.
        All of X is processed as a single batch. This is intended for cases
        when `fit` is not feasible due to very large number of `n_samples`
        or because X is read from a continuous stream.
        The algorithm for incremental mean and std is given in Equation 1.5a,b
        in Chan, Tony F., Gene H. Golub, and Randall J. LeVeque. "Algorithms
        for computing the sample variance: Analysis and recommendations."
        The American Statistician 37.3 (1983): 242-247:
        Parameters
        ----------
        X : {array-like, sparse matrix}, shape [n_samples, n_features]
            The data used to compute the mean and standard deviation
            used for later scaling along the features axis.
        y: Passthrough for ``Pipeline`` compatibility.
        """
        X = check_array(X, accept_sparse=('csr', 'csc'), copy=self.copy,
                        ensure_2d=False, warn_on_dtype=True,
                        estimator=self, dtype=FLOAT_DTYPES)

        if X.ndim == 1:
            warnings.warn(DEPRECATION_MSG_1D, DeprecationWarning)

        # Even in the case of `with_mean=False`, we update the mean anyway
        # This is needed for the incremental computation of the var
        # See incr_mean_variance_axis and _incremental_mean_variance_axis

        if sparse.issparse(X):
            if self.with_mean:
                raise ValueError(
                    "Cannot center sparse matrices: pass `with_mean=False` "
                    "instead. See docstring for motivation and alternatives.")
            if self.with_std:
                # First pass
                if not hasattr(self, 'n_samples_seen_'):
                    self.mean_, self.var_ = mean_variance_axis(X, axis=0)
                    self.n_samples_seen_ = X.shape[0]
                # Next passes
                else:
                    self.mean_, self.var_, self.n_samples_seen_ = \
                        incr_mean_variance_axis(X, axis=0,
                                                last_mean=self.mean_,
                                                last_var=self.var_,
                                                last_n=self.n_samples_seen_)
            else:
                self.mean_ = None
                self.var_ = None
        else:
            # First pass
            if not hasattr(self, 'n_samples_seen_'):
                self.mean_ = .0
                self.n_samples_seen_ = 0
                if self.with_std:
                    self.var_ = .0
                else:
                    self.var_ = None

            self.mean_, self.var_, self.n_samples_seen_ = \
                _incremental_mean_and_var(X, self.mean_, self.var_,
                                          self.n_samples_seen_)

        if self.with_std:
            self.scale_ = _handle_zeros_in_scale(np.sqrt(self.var_))
        else:
            self.scale_ = None

        return self

    def transform(self, X, y=None, copy=None):
        """Perform standardization by centering and scaling
        Parameters
        ----------
        X : array-like, shape [n_samples, n_features]
            The data used to scale along the features axis.
        """
        check_is_fitted(self, 'scale_')

        copy = copy if copy is not None else self.copy
        X = check_array(X, accept_sparse='csr', copy=copy,
                        ensure_2d=False, warn_on_dtype=True,
                        estimator=self, dtype=FLOAT_DTYPES)

        if X.ndim == 1:
            warnings.warn(DEPRECATION_MSG_1D, DeprecationWarning)

        if sparse.issparse(X):
            if self.with_mean:
                raise ValueError(
                    "Cannot center sparse matrices: pass `with_mean=False` "
                    "instead. See docstring for motivation and alternatives.")
            if self.scale_ is not None:
                inplace_column_scale(X, 1 / self.scale_)
        else:
            if self.with_mean:
                X -= self.mean_
            if self.with_std:
                X /= self.scale_
        return X

    def inverse_transform(self, X, copy=None):
        """Scale back the data to the original representation
        Parameters
        ----------
        X : array-like, shape [n_samples, n_features]
            The data used to scale along the features axis.
        """
        check_is_fitted(self, 'scale_')

        copy = copy if copy is not None else self.copy
        if sparse.issparse(X):
            if self.with_mean:
                raise ValueError(
                    "Cannot uncenter sparse matrices: pass `with_mean=False` "
                    "instead See docstring for motivation and alternatives.")
            if not sparse.isspmatrix_csr(X):
                X = X.tocsr()
                copy = False
            if copy:
                X = X.copy()
            if self.scale_ is not None:
                inplace_column_scale(X, self.scale_)
        else:
            X = np.asarray(X)
            if copy:
                X = X.copy()
            if self.with_std:
                X *= self.scale_
            if self.with_mean:
                X += self.mean_
        return X


class MaxAbsScaler(BaseEstimator, TransformerMixin):
    """Scale each feature by its maximum absolute value.
    This estimator scales and translates each feature individually such
    that the maximal absolute value of each feature in the
    training set will be 1.0. It does not shift/center the data, and
    thus does not destroy any sparsity.
    This scaler can also be applied to sparse CSR or CSC matrices.
    .. versionadded:: 0.17
    Parameters
    ----------
    copy : boolean, optional, default is True
        Set to False to perform inplace scaling and avoid a copy (if the input
        is already a numpy array).
    Attributes
    ----------
    scale_ : ndarray, shape (n_features,)
        Per feature relative scaling of the data.
        .. versionadded:: 0.17
           *scale_* attribute.
    max_abs_ : ndarray, shape (n_features,)
        Per feature maximum absolute value.
    n_samples_seen_ : int
        The number of samples processed by the estimator. Will be reset on
        new calls to fit, but increments across ``partial_fit`` calls.
    See also
    --------
    maxabs_scale: Equivalent function without the object oriented API.
    """

    def __init__(self, copy=True):
        self.copy = copy

    def _reset(self):
        """Reset internal data-dependent state of the scaler, if necessary.
        __init__ parameters are not touched.
        """

        # Checking one attribute is enough, becase they are all set together
        # in partial_fit
        if hasattr(self, 'scale_'):
            del self.scale_
            del self.n_samples_seen_
            del self.max_abs_

    def fit(self, X, y=None):
        """Compute the maximum absolute value to be used for later scaling.
        Parameters
        ----------
        X : {array-like, sparse matrix}, shape [n_samples, n_features]
            The data used to compute the per-feature minimum and maximum
            used for later scaling along the features axis.
        """

        # Reset internal state before fitting
        self._reset()
        return self.partial_fit(X, y)

    def partial_fit(self, X, y=None):
        """Online computation of max absolute value of X for later scaling.
        All of X is processed as a single batch. This is intended for cases
        when `fit` is not feasible due to very large number of `n_samples`
        or because X is read from a continuous stream.
        Parameters
        ----------
        X : {array-like, sparse matrix}, shape [n_samples, n_features]
            The data used to compute the mean and standard deviation
            used for later scaling along the features axis.
        y: Passthrough for ``Pipeline`` compatibility.
        """
        X = check_array(X, accept_sparse=('csr', 'csc'), copy=self.copy,
                        ensure_2d=False, estimator=self, dtype=FLOAT_DTYPES)

        if X.ndim == 1:
            warnings.warn(DEPRECATION_MSG_1D, DeprecationWarning)

        if sparse.issparse(X):
            mins, maxs = min_max_axis(X, axis=0)
            max_abs = np.maximum(np.abs(mins), np.abs(maxs))
        else:
            max_abs = np.abs(X).max(axis=0)

        # First pass
        if not hasattr(self, 'n_samples_seen_'):
            self.n_samples_seen_ = X.shape[0]
        # Next passes
        else:
            max_abs = np.maximum(self.max_abs_, max_abs)
            self.n_samples_seen_ += X.shape[0]

        self.max_abs_ = max_abs
        self.scale_ = _handle_zeros_in_scale(max_abs)
        return self

    def transform(self, X, y=None):
        """Scale the data
        Parameters
        ----------
        X : {array-like, sparse matrix}
            The data that should be scaled.
        """
        check_is_fitted(self, 'scale_')
        X = check_array(X, accept_sparse=('csr', 'csc'), copy=self.copy,
                        ensure_2d=False, estimator=self, dtype=FLOAT_DTYPES)

        if X.ndim == 1:
            warnings.warn(DEPRECATION_MSG_1D, DeprecationWarning)

        if sparse.issparse(X):
            inplace_column_scale(X, 1.0 / self.scale_)
        else:
            X /= self.scale_
        return X

    def inverse_transform(self, X):
        """Scale back the data to the original representation
        Parameters
        ----------
        X : {array-like, sparse matrix}
            The data that should be transformed back.
        """
        check_is_fitted(self, 'scale_')
        X = check_array(X, accept_sparse=('csr', 'csc'), copy=self.copy,
                        ensure_2d=False, estimator=self, dtype=FLOAT_DTYPES)
        if X.ndim == 1:
            warnings.warn(DEPRECATION_MSG_1D, DeprecationWarning)

        if sparse.issparse(X):
            inplace_column_scale(X, self.scale_)
        else:
            X *= self.scale_
        return X


def maxabs_scale(X, axis=0, copy=True):
    """Scale each feature to the [-1, 1] range without breaking the sparsity.
    This estimator scales each feature individually such
    that the maximal absolute value of each feature in the
    training set will be 1.0.
    This scaler can also be applied to sparse CSR or CSC matrices.
    Parameters
    ----------
    axis : int (0 by default)
        axis used to scale along. If 0, independently scale each feature,
        otherwise (if 1) scale each sample.
    copy : boolean, optional, default is True
        Set to False to perform inplace scaling and avoid a copy (if the input
        is already a numpy array).
    See also
    --------
    MaxAbsScaler: Performs scaling to the [-1, 1] range using the``Transformer`` API
        (e.g. as part of a preprocessing :class:`sklearn.pipeline.Pipeline`).
    """  # noqa
    # To allow retro-compatibility, we handle here the case of 1D-input
    # From 0.17, 1D-input are deprecated in scaler objects
    # Although, we want to allow the users to keep calling this function
    # with 1D-input.

    # Cast input to array, as we need to check ndim. Prior to 0.17, that was
    # done inside the scaler object fit_transform.
    # If copy is required, it will be done inside the scaler object.
    X = check_array(X, accept_sparse=('csr', 'csc'), copy=False,
                    ensure_2d=False, dtype=FLOAT_DTYPES)
    original_ndim = X.ndim

    if original_ndim == 1:
        X = X.reshape(X.shape[0], 1)

    s = MaxAbsScaler(copy=copy)
    if axis == 0:
        X = s.fit_transform(X)
    else:
        X = s.fit_transform(X.T).T

    if original_ndim == 1:
        X = X.ravel()

    return X


class RobustScaler(BaseEstimator, TransformerMixin):
    """Scale features using statistics that are robust to outliers.
    This Scaler removes the median and scales the data according to
    the quantile range (defaults to IQR: Interquartile Range).
    The IQR is the range between the 1st quartile (25th quantile)
    and the 3rd quartile (75th quantile).
    Centering and scaling happen independently on each feature (or each
    sample, depending on the `axis` argument) by computing the relevant
    statistics on the samples in the training set. Median and  interquartile
    range are then stored to be used on later data using the `transform`
    method.
    Standardization of a dataset is a common requirement for many
    machine learning estimators. Typically this is done by removing the mean
    and scaling to unit variance. However, outliers can often influence the
    sample mean / variance in a negative way. In such cases, the median and
    the interquartile range often give better results.
    .. versionadded:: 0.17
    Read more in the :ref:`User Guide <preprocessing_scaler>`.
    Parameters
    ----------
    with_centering : boolean, True by default
        If True, center the data before scaling.
        This does not work (and will raise an exception) when attempted on
        sparse matrices, because centering them entails building a dense
        matrix which in common use cases is likely to be too large to fit in
        memory.
    with_scaling : boolean, True by default
        If True, scale the data to interquartile range.
    quantile_range : tuple (q_min, q_max), 0.0 < q_min < q_max < 100.0
        Default: (25.0, 75.0) = (1st quantile, 3rd quantile) = IQR
        Quantile range used to calculate ``scale_``.
        .. versionadded:: 0.18
    copy : boolean, optional, default is True
        If False, try to avoid a copy and do inplace scaling instead.
        This is not guaranteed to always work inplace; e.g. if the data is
        not a NumPy array or scipy.sparse CSR matrix, a copy may still be
        returned.
    Attributes
    ----------
    center_ : array of floats
        The median value for each feature in the training set.
    scale_ : array of floats
        The (scaled) interquartile range for each feature in the training set.
        .. versionadded:: 0.17
           *scale_* attribute.
    See also
    --------
    robust_scale: Equivalent function without the object oriented API.
    :class:`sklearn.decomposition.PCA`
        Further removes the linear correlation across features with
        'whiten=True'.
    Notes
    -----
    See examples/preprocessing/plot_robust_scaling.py for an example.
    https://en.wikipedia.org/wiki/Median_(statistics)
    https://en.wikipedia.org/wiki/Interquartile_range
    """

    def __init__(self, with_centering=True, with_scaling=True,
                 quantile_range=(25.0, 75.0), copy=True):
        self.with_centering = with_centering
        self.with_scaling = with_scaling
        self.quantile_range = quantile_range
        self.copy = copy

    def _check_array(self, X, copy):
        """Makes sure centering is not enabled for sparse matrices."""
        X = check_array(X, accept_sparse=('csr', 'csc'), copy=self.copy,
                        ensure_2d=False, estimator=self, dtype=FLOAT_DTYPES)

        if X.ndim == 1:
            warnings.warn(DEPRECATION_MSG_1D, DeprecationWarning)

        if sparse.issparse(X):
            if self.with_centering:
                raise ValueError(
                    "Cannot center sparse matrices: use `with_centering=False`"
                    " instead. See docstring for motivation and alternatives.")
        return X

    def fit(self, X, y=None):
        """Compute the median and quantiles to be used for scaling.
        Parameters
        ----------
        X : array-like, shape [n_samples, n_features]
            The data used to compute the median and quantiles
            used for later scaling along the features axis.
        """
        if sparse.issparse(X):
            raise TypeError("RobustScaler cannot be fitted on sparse inputs")
        X = self._check_array(X, self.copy)
        if X.ndim == 1:
            warnings.warn(DEPRECATION_MSG_1D, DeprecationWarning)
        if self.with_centering:
            self.center_ = np.median(X, axis=0)

        if self.with_scaling:
            q_min, q_max = self.quantile_range
            if not 0 <= q_min <= q_max <= 100:
                raise ValueError("Invalid quantile range: %s" %
                                 str(self.quantile_range))

            q = np.percentile(X, self.quantile_range, axis=0)
            self.scale_ = (q[1] - q[0])
            self.scale_ = _handle_zeros_in_scale(self.scale_, copy=False)
        return self

    def transform(self, X, y=None):
        """Center and scale the data
        Parameters
        ----------
        X : array-like
            The data used to scale along the specified axis.
        """
        if self.with_centering:
            check_is_fitted(self, 'center_')
        if self.with_scaling:
            check_is_fitted(self, 'scale_')
        X = self._check_array(X, self.copy)
        if X.ndim == 1:
            warnings.warn(DEPRECATION_MSG_1D, DeprecationWarning)

        if sparse.issparse(X):
            if self.with_scaling:
                inplace_column_scale(X, 1.0 / self.scale_)
        else:
            if self.with_centering:
                X -= self.center_
            if self.with_scaling:
                X /= self.scale_
        return X

    def inverse_transform(self, X):
        """Scale back the data to the original representation
        Parameters
        ----------
        X : array-like
            The data used to scale along the specified axis.
        """
        if self.with_centering:
            check_is_fitted(self, 'center_')
        if self.with_scaling:
            check_is_fitted(self, 'scale_')
        X = self._check_array(X, self.copy)
        if X.ndim == 1:
            warnings.warn(DEPRECATION_MSG_1D, DeprecationWarning)

        if sparse.issparse(X):
            if self.with_scaling:
                inplace_column_scale(X, self.scale_)
        else:
            if self.with_scaling:
                X *= self.scale_
            if self.with_centering:
                X += self.center_
        return X


def robust_scale(X, axis=0, with_centering=True, with_scaling=True,
                 quantile_range=(25.0, 75.0), copy=True):
    """Standardize a dataset along any axis
    Center to the median and component wise scale
    according to the interquartile range.
    Read more in the :ref:`User Guide <preprocessing_scaler>`.
    Parameters
    ----------
    X : array-like
        The data to center and scale.
    axis : int (0 by default)
        axis used to compute the medians and IQR along. If 0,
        independently scale each feature, otherwise (if 1) scale
        each sample.
    with_centering : boolean, True by default
        If True, center the data before scaling.
    with_scaling : boolean, True by default
        If True, scale the data to unit variance (or equivalently,
        unit standard deviation).
    quantile_range : tuple (q_min, q_max), 0.0 < q_min < q_max < 100.0
        Default: (25.0, 75.0) = (1st quantile, 3rd quantile) = IQR
        Quantile range used to calculate ``scale_``.
        .. versionadded:: 0.18
    copy : boolean, optional, default is True
        set to False to perform inplace row normalization and avoid a
        copy (if the input is already a numpy array or a scipy.sparse
        CSR matrix and if axis is 1).
    Notes
    -----
    This implementation will refuse to center scipy.sparse matrices
    since it would make them non-sparse and would potentially crash the
    program with memory exhaustion problems.
    Instead the caller is expected to either set explicitly
    `with_centering=False` (in that case, only variance scaling will be
    performed on the features of the CSR matrix) or to call `X.toarray()`
    if he/she expects the materialized dense array to fit in memory.
    To avoid memory copy the caller should pass a CSR matrix.
    See also
    --------
    RobustScaler: Performs centering and scaling using the ``Transformer`` API
        (e.g. as part of a preprocessing :class:`sklearn.pipeline.Pipeline`).
    """
    s = RobustScaler(with_centering=with_centering, with_scaling=with_scaling,
                     quantile_range=quantile_range, copy=copy)
    if axis == 0:
        return s.fit_transform(X)
    else:
        return s.fit_transform(X.T).T


class PolynomialFeatures(BaseEstimator, TransformerMixin):
    """Generate polynomial and interaction features.
    Generate a new feature matrix consisting of all polynomial combinations
    of the features with degree less than or equal to the specified degree.
    For example, if an input sample is two dimensional and of the form
    [a, b], the degree-2 polynomial features are [1, a, b, a^2, ab, b^2].
    Parameters
    ----------
    degree : integer
        The degree of the polynomial features. Default = 2.
    interaction_only : boolean, default = False
        If true, only interaction features are produced: features that are
        products of at most ``degree`` *distinct* input features (so not
        ``x[1] ** 2``, ``x[0] * x[2] ** 3``, etc.).
    include_bias : boolean
        If True (default), then include a bias column, the feature in which
        all polynomial powers are zero (i.e. a column of ones - acts as an
        intercept term in a linear model).
    Examples
    --------
    >>> X = np.arange(6).reshape(3, 2)
    >>> X
    array([[0, 1],
           [2, 3],
           [4, 5]])
    >>> poly = PolynomialFeatures(2)
    >>> poly.fit_transform(X)
    array([[  1.,   0.,   1.,   0.,   0.,   1.],
           [  1.,   2.,   3.,   4.,   6.,   9.],
           [  1.,   4.,   5.,  16.,  20.,  25.]])
    >>> poly = PolynomialFeatures(interaction_only=True)
    >>> poly.fit_transform(X)
    array([[  1.,   0.,   1.,   0.],
           [  1.,   2.,   3.,   6.],
           [  1.,   4.,   5.,  20.]])
    Attributes
    ----------
    powers_ : array, shape (n_output_features, n_input_features)
        powers_[i, j] is the exponent of the jth input in the ith output.
    n_input_features_ : int
        The total number of input features.
    n_output_features_ : int
        The total number of polynomial output features. The number of output
        features is computed by iterating over all suitably sized combinations
        of input features.
    Notes
    -----
    Be aware that the number of features in the output array scales
    polynomially in the number of features of the input array, and
    exponentially in the degree. High degrees can cause overfitting.
    See :ref:`examples/linear_model/plot_polynomial_interpolation.py
    <sphx_glr_auto_examples_linear_model_plot_polynomial_interpolation.py>`
    """
    def __init__(self, degree=2, interaction_only=False, include_bias=True,symmetry_list=None):
        self.degree = degree
        self.interaction_only = interaction_only
        self.include_bias = include_bias
        self.symmetry_list = symmetry_list

    @staticmethod
    def _combinations(self,n_features, degree, interaction_only, include_bias):
        comb = (combinations if interaction_only else combinations_w_r)
        start = int(not include_bias)
        combos = chain.from_iterable(comb(range(n_features), i)
                                   for i in range(start, degree + 1))

        # Select items with only +1 symmetry
        combos_ok = []
        for line in combos:
#            print "Internal", line  # tuple
            signature = np.prod( np.array(self.symmetry_list)[list(line)] )
            if signature >=0:
                combos_ok.append(line)
#            signature = np.prod(np.power(np.array(self.symmetry_list)),line)
#            if signature >=0:
#                combos.append(line)
        return combos_ok

    @property
    def powers_(self):
        check_is_fitted(self, 'n_input_features_')

        combinations = self._combinations(self,self.n_input_features_, self.degree,
                                          self.interaction_only,
                                          self.include_bias)
        return np.vstack(bincount(c, minlength=self.n_input_features_)
                         for c in combinations)

    def get_feature_names(self, input_features=None):
        """
        Return feature names for output features
        Parameters
        ----------
        input_features : list of string, length n_features, optional
            String names for input features if available. By default,
            "x0", "x1", ... "xn_features" is used.
        Returns
        -------
        output_feature_names : list of string, length n_output_features
        """
        powers = self.powers_
        if input_features is None:
            input_features = ['x%d' % i for i in range(powers.shape[1])]
        feature_names = []
        for row in powers:
            inds = np.where(row)[0]
            if len(inds):
                name = " ".join("%s^%d" % (input_features[ind], exp)
                                if exp != 1 else input_features[ind]
                                for ind, exp in zip(inds, row[inds]))
            else:
                name = "1"
            feature_names.append(name)
        return feature_names

    def fit(self, X, y=None):
        """
        Compute number of output features.
        """
        n_samples, n_features = check_array(X).shape
        combinations = self._combinations(self,n_features, self.degree,
                                          self.interaction_only,
                                          self.include_bias)
        self.n_input_features_ = n_features
        self.n_output_features_ = sum(1 for _ in combinations)
        return self

    def transform(self, X, y=None):
        """Transform data to polynomial features
        Parameters
        ----------
        X : array-like, shape [n_samples, n_features]
            The data to transform, row by row.
        Returns
        -------
        XP : np.ndarray shape [n_samples, NP]
            The matrix of features, where NP is the number of polynomial
            features generated from the combination of inputs.
        """
        check_is_fitted(self, ['n_input_features_', 'n_output_features_'])

        X = check_array(X, dtype=FLOAT_DTYPES)
        n_samples, n_features = X.shape

        if n_features != self.n_input_features_:
            raise ValueError("X shape does not match training shape")

        # allocate output data
        XP = np.empty((n_samples, self.n_output_features_), dtype=X.dtype)

        combinations = self._combinations(self,n_features, self.degree,
                                          self.interaction_only,
                                          self.include_bias)
        for i, c in enumerate(combinations):
            XP[:, i] = X[:, c].prod(1)

        return XP


def normalize(X, norm='l2', axis=1, copy=True, return_norm=False):
    """Scale input vectors individually to unit norm (vector length).
    Read more in the :ref:`User Guide <preprocessing_normalization>`.
    Parameters
    ----------
    X : {array-like, sparse matrix}, shape [n_samples, n_features]
        The data to normalize, element by element.
        scipy.sparse matrices should be in CSR format to avoid an
        un-necessary copy.
    norm : 'l1', 'l2', or 'max', optional ('l2' by default)
        The norm to use to normalize each non zero sample (or each non-zero
        feature if axis is 0).
    axis : 0 or 1, optional (1 by default)
        axis used to normalize the data along. If 1, independently normalize
        each sample, otherwise (if 0) normalize each feature.
    copy : boolean, optional, default True
        set to False to perform inplace row normalization and avoid a
        copy (if the input is already a numpy array or a scipy.sparse
        CSR matrix and if axis is 1).
    return_norm : boolean, default False
        whether to return the computed norms
    See also
    --------
    Normalizer: Performs normalization using the ``Transformer`` API
        (e.g. as part of a preprocessing :class:`sklearn.pipeline.Pipeline`).
    """
    if norm not in ('l1', 'l2', 'max'):
        raise ValueError("'%s' is not a supported norm" % norm)

    if axis == 0:
        sparse_format = 'csc'
    elif axis == 1:
        sparse_format = 'csr'
    else:
        raise ValueError("'%d' is not a supported axis" % axis)

    X = check_array(X, sparse_format, copy=copy, warn_on_dtype=True,
                    estimator='the normalize function', dtype=FLOAT_DTYPES)
    if axis == 0:
        X = X.T

    if sparse.issparse(X):
        if norm == 'l1':
            inplace_csr_row_normalize_l1(X)
        elif norm == 'l2':
            inplace_csr_row_normalize_l2(X)
        elif norm == 'max':
            _, norms = min_max_axis(X, 1)
            norms = norms.repeat(np.diff(X.indptr))
            mask = norms != 0
            X.data[mask] /= norms[mask]
    else:
        if norm == 'l1':
            norms = np.abs(X).sum(axis=1)
        elif norm == 'l2':
            norms = row_norms(X)
        elif norm == 'max':
            norms = np.max(X, axis=1)
        norms = _handle_zeros_in_scale(norms, copy=False)
        X /= norms[:, np.newaxis]

    if axis == 0:
        X = X.T

    if return_norm:
        return X, norms
    else:
        return X


class Normalizer(BaseEstimator, TransformerMixin):
    """Normalize samples individually to unit norm.
    Each sample (i.e. each row of the data matrix) with at least one
    non zero component is rescaled independently of other samples so
    that its norm (l1 or l2) equals one.
    This transformer is able to work both with dense numpy arrays and
    scipy.sparse matrix (use CSR format if you want to avoid the burden of
    a copy / conversion).
    Scaling inputs to unit norms is a common operation for text
    classification or clustering for instance. For instance the dot
    product of two l2-normalized TF-IDF vectors is the cosine similarity
    of the vectors and is the base similarity metric for the Vector
    Space Model commonly used by the Information Retrieval community.
    Read more in the :ref:`User Guide <preprocessing_normalization>`.
    Parameters
    ----------
    norm : 'l1', 'l2', or 'max', optional ('l2' by default)
        The norm to use to normalize each non zero sample.
    copy : boolean, optional, default True
        set to False to perform inplace row normalization and avoid a
        copy (if the input is already a numpy array or a scipy.sparse
        CSR matrix).
    Notes
    -----
    This estimator is stateless (besides constructor parameters), the
    fit method does nothing but is useful when used in a pipeline.
    See also
    --------
    normalize: Equivalent function without the object oriented API.
    """

    def __init__(self, norm='l2', copy=True):
        self.norm = norm
        self.copy = copy

    def fit(self, X, y=None):
        """Do nothing and return the estimator unchanged
        This method is just there to implement the usual API and hence
        work in pipelines.
        """
        X = check_array(X, accept_sparse='csr')
        return self

    def transform(self, X, y=None, copy=None):
        """Scale each non zero row of X to unit norm
        Parameters
        ----------
        X : {array-like, sparse matrix}, shape [n_samples, n_features]
            The data to normalize, row by row. scipy.sparse matrices should be
            in CSR format to avoid an un-necessary copy.
        """
        copy = copy if copy is not None else self.copy
        X = check_array(X, accept_sparse='csr')
        return normalize(X, norm=self.norm, axis=1, copy=copy)


def binarize(X, threshold=0.0, copy=True):
    """Boolean thresholding of array-like or scipy.sparse matrix
    Read more in the :ref:`User Guide <preprocessing_binarization>`.
    Parameters
    ----------
    X : {array-like, sparse matrix}, shape [n_samples, n_features]
        The data to binarize, element by element.
        scipy.sparse matrices should be in CSR or CSC format to avoid an
        un-necessary copy.
    threshold : float, optional (0.0 by default)
        Feature values below or equal to this are replaced by 0, above it by 1.
        Threshold may not be less than 0 for operations on sparse matrices.
    copy : boolean, optional, default True
        set to False to perform inplace binarization and avoid a copy
        (if the input is already a numpy array or a scipy.sparse CSR / CSC
        matrix and if axis is 1).
    See also
    --------
    Binarizer: Performs binarization using the ``Transformer`` API
        (e.g. as part of a preprocessing :class:`sklearn.pipeline.Pipeline`).
    """
    X = check_array(X, accept_sparse=['csr', 'csc'], copy=copy)
    if sparse.issparse(X):
        if threshold < 0:
            raise ValueError('Cannot binarize a sparse matrix with threshold '
                             '< 0')
        cond = X.data > threshold
        not_cond = np.logical_not(cond)
        X.data[cond] = 1
        X.data[not_cond] = 0
        X.eliminate_zeros()
    else:
        cond = X > threshold
        not_cond = np.logical_not(cond)
        X[cond] = 1
        X[not_cond] = 0
    return X


class Binarizer(BaseEstimator, TransformerMixin):
    """Binarize data (set feature values to 0 or 1) according to a threshold
    Values greater than the threshold map to 1, while values less than
    or equal to the threshold map to 0. With the default threshold of 0,
    only positive values map to 1.
    Binarization is a common operation on text count data where the
    analyst can decide to only consider the presence or absence of a
    feature rather than a quantified number of occurrences for instance.
    It can also be used as a pre-processing step for estimators that
    consider boolean random variables (e.g. modelled using the Bernoulli
    distribution in a Bayesian setting).
    Read more in the :ref:`User Guide <preprocessing_binarization>`.
    Parameters
    ----------
    threshold : float, optional (0.0 by default)
        Feature values below or equal to this are replaced by 0, above it by 1.
        Threshold may not be less than 0 for operations on sparse matrices.
    copy : boolean, optional, default True
        set to False to perform inplace binarization and avoid a copy (if
        the input is already a numpy array or a scipy.sparse CSR matrix).
    Notes
    -----
    If the input is a sparse matrix, only the non-zero values are subject
    to update by the Binarizer class.
    This estimator is stateless (besides constructor parameters), the
    fit method does nothing but is useful when used in a pipeline.
    See also
    --------
    binarize: Equivalent function without the object oriented API.
    """

    def __init__(self, threshold=0.0, copy=True):
        self.threshold = threshold
        self.copy = copy

    def fit(self, X, y=None):
        """Do nothing and return the estimator unchanged
        This method is just there to implement the usual API and hence
        work in pipelines.
        """
        check_array(X, accept_sparse='csr')
        return self

    def transform(self, X, y=None, copy=None):
        """Binarize each element of X
        Parameters
        ----------
        X : {array-like, sparse matrix}, shape [n_samples, n_features]
            The data to binarize, element by element.
            scipy.sparse matrices should be in CSR format to avoid an
            un-necessary copy.
        """
        copy = copy if copy is not None else self.copy
        return binarize(X, threshold=self.threshold, copy=copy)


class KernelCenterer(BaseEstimator, TransformerMixin):
    """Center a kernel matrix
    Let K(x, z) be a kernel defined by phi(x)^T phi(z), where phi is a
    function mapping x to a Hilbert space. KernelCenterer centers (i.e.,
    normalize to have zero mean) the data without explicitly computing phi(x).
    It is equivalent to centering phi(x) with
    sklearn.preprocessing.StandardScaler(with_std=False).
    Read more in the :ref:`User Guide <kernel_centering>`.
    """

    def fit(self, K, y=None):
        """Fit KernelCenterer
        Parameters
        ----------
        K : numpy array of shape [n_samples, n_samples]
            Kernel matrix.
        Returns
        -------
        self : returns an instance of self.
        """
        K = check_array(K, dtype=FLOAT_DTYPES)
        n_samples = K.shape[0]
        self.K_fit_rows_ = np.sum(K, axis=0) / n_samples
        self.K_fit_all_ = self.K_fit_rows_.sum() / n_samples
        return self

    def transform(self, K, y=None, copy=True):
        """Center kernel matrix.
        Parameters
        ----------
        K : numpy array of shape [n_samples1, n_samples2]
            Kernel matrix.
        copy : boolean, optional, default True
            Set to False to perform inplace computation.
        Returns
        -------
        K_new : numpy array of shape [n_samples1, n_samples2]
        """
        check_is_fitted(self, 'K_fit_all_')

        K = check_array(K, copy=copy, dtype=FLOAT_DTYPES)

        K_pred_cols = (np.sum(K, axis=1) /
                       self.K_fit_rows_.shape[0])[:, np.newaxis]

        K -= self.K_fit_rows_
        K -= K_pred_cols
        K += self.K_fit_all_

        return K

    @property
    def _pairwise(self):
        return True


def add_dummy_feature(X, value=1.0):
    """Augment dataset with an additional dummy feature.
    This is useful for fitting an intercept term with implementations which
    cannot otherwise fit it directly.
    Parameters
    ----------
    X : {array-like, sparse matrix}, shape [n_samples, n_features]
        Data.
    value : float
        Value to use for the dummy feature.
    Returns
    -------
    X : {array, sparse matrix}, shape [n_samples, n_features + 1]
        Same data with dummy feature added as first column.
    Examples
    --------
    >>> from sklearn.preprocessing import add_dummy_feature
    >>> add_dummy_feature([[0, 1], [1, 0]])
    array([[ 1.,  0.,  1.],
           [ 1.,  1.,  0.]])
    """
    X = check_array(X, accept_sparse=['csc', 'csr', 'coo'], dtype=FLOAT_DTYPES)
    n_samples, n_features = X.shape
    shape = (n_samples, n_features + 1)
    if sparse.issparse(X):
        if sparse.isspmatrix_coo(X):
            # Shift columns to the right.
            col = X.col + 1
            # Column indices of dummy feature are 0 everywhere.
            col = np.concatenate((np.zeros(n_samples), col))
            # Row indices of dummy feature are 0, ..., n_samples-1.
            row = np.concatenate((np.arange(n_samples), X.row))
            # Prepend the dummy feature n_samples times.
            data = np.concatenate((np.ones(n_samples) * value, X.data))
            return sparse.coo_matrix((data, (row, col)), shape)
        elif sparse.isspmatrix_csc(X):
            # Shift index pointers since we need to add n_samples elements.
            indptr = X.indptr + n_samples
            # indptr[0] must be 0.
            indptr = np.concatenate((np.array([0]), indptr))
            # Row indices of dummy feature are 0, ..., n_samples-1.
            indices = np.concatenate((np.arange(n_samples), X.indices))
            # Prepend the dummy feature n_samples times.
            data = np.concatenate((np.ones(n_samples) * value, X.data))
            return sparse.csc_matrix((data, indices, indptr), shape)
        else:
            klass = X.__class__
            return klass(add_dummy_feature(X.tocoo(), value))
    else:
        return np.hstack((np.ones((n_samples, 1)) * value, X))


def _transform_selected(X, transform, selected="all", copy=True):
    """Apply a transform function to portion of selected features
    Parameters
    ----------
    X : {array-like, sparse matrix}, shape [n_samples, n_features]
        Dense array or sparse matrix.
    transform : callable
        A callable transform(X) -> X_transformed
    copy : boolean, optional
        Copy X even if it could be avoided.
    selected: "all" or array of indices or mask
        Specify which features to apply the transform to.
    Returns
    -------
    X : array or sparse matrix, shape=(n_samples, n_features_new)
    """
    X = check_array(X, accept_sparse='csc', copy=copy, dtype=FLOAT_DTYPES)

    if isinstance(selected, six.string_types) and selected == "all":
        return transform(X)

    if len(selected) == 0:
        return X

    n_features = X.shape[1]
    ind = np.arange(n_features)
    sel = np.zeros(n_features, dtype=bool)
    sel[np.asarray(selected)] = True
    not_sel = np.logical_not(sel)
    n_selected = np.sum(sel)

    if n_selected == 0:
        # No features selected.
        return X
    elif n_selected == n_features:
        # All features selected.
        return transform(X)
    else:
        X_sel = transform(X[:, ind[sel]])
        X_not_sel = X[:, ind[not_sel]]

        if sparse.issparse(X_sel) or sparse.issparse(X_not_sel):
            return sparse.hstack((X_sel, X_not_sel))
        else:
            return np.hstack((X_sel, X_not_sel))


class OneHotEncoder(BaseEstimator, TransformerMixin):
    """Encode categorical integer features using a one-hot aka one-of-K scheme.
    The input to this transformer should be a matrix of integers, denoting
    the values taken on by categorical (discrete) features. The output will be
    a sparse matrix where each column corresponds to one possible value of one
    feature. It is assumed that input features take on values in the range
    [0, n_values).
    This encoding is needed for feeding categorical data to many scikit-learn
    estimators, notably linear models and SVMs with the standard kernels.
    Note: a one-hot encoding of y labels should use a LabelBinarizer
    instead.
    Read more in the :ref:`User Guide <preprocessing_categorical_features>`.
    Parameters
    ----------
    n_values : 'auto', int or array of ints
        Number of values per feature.
        - 'auto' : determine value range from training data.
        - int : number of categorical values per feature.
                Each feature value should be in ``range(n_values)``
        - array : ``n_values[i]`` is the number of categorical values in
                  ``X[:, i]``. Each feature value should be
                  in ``range(n_values[i])``
    categorical_features : "all" or array of indices or mask
        Specify what features are treated as categorical.
        - 'all' (default): All features are treated as categorical.
        - array of indices: Array of categorical feature indices.
        - mask: Array of length n_features and with dtype=bool.
        Non-categorical features are always stacked to the right of the matrix.
    dtype : number type, default=np.float
        Desired dtype of output.
    sparse : boolean, default=True
        Will return sparse matrix if set True else will return an array.
    handle_unknown : str, 'error' or 'ignore'
        Whether to raise an error or ignore if a unknown categorical feature is
        present during transform.
    Attributes
    ----------
    active_features_ : array
        Indices for active features, meaning values that actually occur
        in the training set. Only available when n_values is ``'auto'``.
    feature_indices_ : array of shape (n_features,)
        Indices to feature ranges.
        Feature ``i`` in the original data is mapped to features
        from ``feature_indices_[i]`` to ``feature_indices_[i+1]``
        (and then potentially masked by `active_features_` afterwards)
    n_values_ : array of shape (n_features,)
        Maximum number of values per feature.
    Examples
    --------
    Given a dataset with three features and two samples, we let the encoder
    find the maximum value per feature and transform the data to a binary
    one-hot encoding.
    >>> from sklearn.preprocessing import OneHotEncoder
    >>> enc = OneHotEncoder()
    >>> enc.fit([[0, 0, 3], [1, 1, 0], [0, 2, 1], \
[1, 0, 2]])  # doctest: +ELLIPSIS
    OneHotEncoder(categorical_features='all', dtype=<... 'numpy.float64'>,
           handle_unknown='error', n_values='auto', sparse=True)
    >>> enc.n_values_
    array([2, 3, 4])
    >>> enc.feature_indices_
    array([0, 2, 5, 9])
    >>> enc.transform([[0, 1, 1]]).toarray()
    array([[ 1.,  0.,  0.,  1.,  0.,  0.,  1.,  0.,  0.]])
    See also
    --------
    sklearn.feature_extraction.DictVectorizer : performs a one-hot encoding of
      dictionary items (also handles string-valued features).
    sklearn.feature_extraction.FeatureHasher : performs an approximate one-hot
      encoding of dictionary items or strings.
    sklearn.preprocessing.LabelBinarizer : binarizes labels in a one-vs-all
      fashion.
    sklearn.preprocessing.MultiLabelBinarizer : transforms between iterable of
      iterables and a multilabel format, e.g. a (samples x classes) binary
      matrix indicating the presence of a class label.
    sklearn.preprocessing.LabelEncoder : encodes labels with values between 0
      and n_classes-1.
    """
    def __init__(self, n_values="auto", categorical_features="all",
                 dtype=np.float64, sparse=True, handle_unknown='error'):
        self.n_values = n_values
        self.categorical_features = categorical_features
        self.dtype = dtype
        self.sparse = sparse
        self.handle_unknown = handle_unknown

    def fit(self, X, y=None):
        """Fit OneHotEncoder to X.
        Parameters
        ----------
        X : array-like, shape [n_samples, n_feature]
            Input array of type int.
        Returns
        -------
        self
        """
        self.fit_transform(X)
        return self

    def _fit_transform(self, X):
        """Assumes X contains only categorical features."""
        X = check_array(X, dtype=np.int)
        if np.any(X < 0):
            raise ValueError("X needs to contain only non-negative integers.")
        n_samples, n_features = X.shape
        if (isinstance(self.n_values, six.string_types) and
                self.n_values == 'auto'):
            n_values = np.max(X, axis=0) + 1
        elif isinstance(self.n_values, numbers.Integral):
            if (np.max(X, axis=0) >= self.n_values).any():
                raise ValueError("Feature out of bounds for n_values=%d"
                                 % self.n_values)
            n_values = np.empty(n_features, dtype=np.int)
            n_values.fill(self.n_values)
        else:
            try:
                n_values = np.asarray(self.n_values, dtype=int)
            except (ValueError, TypeError):
                raise TypeError("Wrong type for parameter `n_values`. Expected"
                                " 'auto', int or array of ints, got %r"
                                % type(X))
            if n_values.ndim < 1 or n_values.shape[0] != X.shape[1]:
                raise ValueError("Shape mismatch: if n_values is an array,"
                                 " it has to be of shape (n_features,).")

        self.n_values_ = n_values
        n_values = np.hstack([[0], n_values])
        indices = np.cumsum(n_values)
        self.feature_indices_ = indices

        column_indices = (X + indices[:-1]).ravel()
        row_indices = np.repeat(np.arange(n_samples, dtype=np.int32),
                                n_features)
        data = np.ones(n_samples * n_features)
        out = sparse.coo_matrix((data, (row_indices, column_indices)),
                                shape=(n_samples, indices[-1]),
                                dtype=self.dtype).tocsr()

        if (isinstance(self.n_values, six.string_types) and
                self.n_values == 'auto'):
            mask = np.array(out.sum(axis=0)).ravel() != 0
            active_features = np.where(mask)[0]
            out = out[:, active_features]
            self.active_features_ = active_features

        return out if self.sparse else out.toarray()

    def fit_transform(self, X, y=None):
        """Fit OneHotEncoder to X, then transform X.
        Equivalent to self.fit(X).transform(X), but more convenient and more
        efficient. See fit for the parameters, transform for the return value.
        """
        return _transform_selected(X, self._fit_transform,
                                   self.categorical_features, copy=True)

    def _transform(self, X):
        """Assumes X contains only categorical features."""
        X = check_array(X, dtype=np.int)
        if np.any(X < 0):
            raise ValueError("X needs to contain only non-negative integers.")
        n_samples, n_features = X.shape

        indices = self.feature_indices_
        if n_features != indices.shape[0] - 1:
            raise ValueError("X has different shape than during fitting."
                             " Expected %d, got %d."
                             % (indices.shape[0] - 1, n_features))

        # We use only those categorical features of X that are known using fit.
        # i.e lesser than n_values_ using mask.
        # This means, if self.handle_unknown is "ignore", the row_indices and
        # col_indices corresponding to the unknown categorical feature are
        # ignored.
        mask = (X < self.n_values_).ravel()
        if np.any(~mask):
            if self.handle_unknown not in ['error', 'ignore']:
                raise ValueError("handle_unknown should be either error or "
                                 "unknown got %s" % self.handle_unknown)
            if self.handle_unknown == 'error':
                raise ValueError("unknown categorical feature present %s "
                                 "during transform." % X.ravel()[~mask])

        column_indices = (X + indices[:-1]).ravel()[mask]
        row_indices = np.repeat(np.arange(n_samples, dtype=np.int32),
                                n_features)[mask]
        data = np.ones(np.sum(mask))
        out = sparse.coo_matrix((data, (row_indices, column_indices)),
                                shape=(n_samples, indices[-1]),
                                dtype=self.dtype).tocsr()
        if (isinstance(self.n_values, six.string_types) and
                self.n_values == 'auto'):
            out = out[:, self.active_features_]

        return out if self.sparse else out.toarray()

    def transform(self, X):
        """Transform X using one-hot encoding.
        Parameters
        ----------
        X : array-like, shape [n_samples, n_features]
            Input array of type int.
        Returns
        -------
        X_out : sparse matrix if sparse=True else a 2-d array, dtype=int
            Transformed input.
        """
        return _transform_selected(X, self._transform,
                                   self.categorical_features, copy=True)



"""
Author: Daniel Wysocki (daniel.m.wysocki@gmail.com)
"""
from __future__ import print_function

import six
import math
import numpy as np

try:
    import cupy
    xpy_default = cupy
    cupy_here = True
    junk_to_check_installed = cupy.array(5)  # this will fail if GPU not installed correctly
except: # ImportError:
    print(" no cupy")
    cupy_here = False
    import numpy as cupy
    xpy_default = np


# Many coefficients include division by powers of two. Multiplication is
# cheaper, and reciprocals of powers of two can be written exactly, so we just
# multiply by inverses.  Reciprocal powers of two greater than eight are given
# names for clarity, as only 1/2 = 0.5 and 1/4 = 0.25 can be expected to be
# understood immediately.
_one_by_eight = 0.125
_one_by_sixteen = 0.0625
_one_by_thirtytwo = 0.03125
_one_by_sixtyfour = 0.015625
_one_by_onetwentyeight = 0.0078125
_one_by_twofiftysix = 0.00390625
_one_by_fivetwelve = 0.001953125
_one_by_fortyninetysix = 0.000244140625

# Coefficients that appear in the various spherical harmonics.
_coeffs_np = {
    2 : {
        -2: math.sqrt(5.0 / (64.0 * math.pi)),
        -1: math.sqrt(5.0 / (16.0 * math.pi)),
        0: math.sqrt(15.0 / (32.0 * math.pi)),
        +1: math.sqrt(5.0 / (16.0 * math.pi)),
        +2: math.sqrt(5.0 / (64.0 * math.pi)),
    },
    3 : {
        -3: math.sqrt(21.0 / (2.0 * math.pi)),
        -2: math.sqrt(7.0 / (4.0 * math.pi)),
        -1: math.sqrt(35.0 / (2.0 * math.pi)) * _one_by_thirtytwo,
        0: math.sqrt(105.0 / (2.0 * math.pi)) * 0.25,
        +1: -math.sqrt(35.0 / (2.0 * math.pi)) * _one_by_thirtytwo,
        +2: math.sqrt(7.0 / math.pi) * 0.5,
        +3: -math.sqrt(21.0 / (2.0 * math.pi)),
    },
    4: {
        -4: 3.0 * math.sqrt(7.0 / math.pi),
        -3: 3.0 * math.sqrt(7.0 / (2.0 * math.pi)),
        -2: 3.0 / math.sqrt(math.pi) * 0.25,
        -1: 3.0 / math.sqrt(2.0 * math.pi) * _one_by_thirtytwo,
        0: 3.0 * math.sqrt(5.0 / (2.0 * math.pi)) * _one_by_sixteen,
        +1: 3.0 / math.sqrt(2.0*math.pi) * _one_by_thirtytwo,
        +2: 3.0 / math.sqrt(math.pi) * 0.25,
        +3: -3.0 * math.sqrt(7.0 / (2.0 * math.pi)),
        +4: 3.0 * math.sqrt(7.0 / math.pi),
    },
    5: {
        -5: math.sqrt(330.0 / math.pi),
        -4: math.sqrt(33.0 / math.pi),
        -3: math.sqrt(33.0 / (2.0 * math.pi)) * 0.25,
        -2: math.sqrt(11.0 / math.pi) * _one_by_eight,
        -1: math.sqrt(77.0 / math.pi) * _one_by_twofiftysix,
        0: math.sqrt(1155.0 / (2.0 * math.pi)) * _one_by_thirtytwo,
        +1: math.sqrt(77.0 / math.pi) * _one_by_twofiftysix,
        +2: math.sqrt(11.0 / math.pi) * _one_by_eight,
        +3: -math.sqrt(33.0 / (2.0 * math.pi)) * 0.25,
        +4: math.sqrt(33.0 / math.pi),
        +5: -math.sqrt(330.0 / math.pi),
    },
    6: {
        -6: 3.0 * math.sqrt(715.0 / math.pi) * 0.5,
        -5: math.sqrt(2145.0 / math.pi) * 0.5,
        -4: math.sqrt(195.0 / (2.0 * math.pi)) * _one_by_eight,
        -3: 3.0 * math.sqrt(13.0 / math.pi) * _one_by_thirtytwo,
        -2: math.sqrt(13.0 / math.pi) * _one_by_twofiftysix,
        -1: math.sqrt(65.0 / (2.0 * math.pi)) * _one_by_sixtyfour,
        0: math.sqrt(1365.0 / math.pi) * _one_by_fivetwelve,
        +1: math.sqrt(65.0 / (2.0 * math.pi)) * _one_by_sixtyfour,
        +2: math.sqrt(13.0 / math.pi) * _one_by_twofiftysix,
        +3: -3.0 * math.sqrt(13.0 / math.pi) * _one_by_thirtytwo,
        +4: math.sqrt(195.0 / (2.0 * math.pi)) * _one_by_eight,
        +5: -math.sqrt(2145.0 / math.pi) * 0.5,
        +6: 3.0 * math.sqrt(715.0 / math.pi) * 0.5,
    },
    7: {
        -7: math.sqrt(15015.0 / (2.0 * math.pi)),
        -6: math.sqrt(2145.0 / math.pi) * 0.5,
        -5: math.sqrt(165.0 / (2.0 * math.pi)) * _one_by_eight,
        -4: math.sqrt(165.0 / (2.0 * math.pi)) * _one_by_sixteen,
        -3: math.sqrt(15.0 / (2.0 * math.pi)) * _one_by_onetwentyeight,
        -2: math.sqrt(15.0 / math.pi) * _one_by_fivetwelve,
        -1: 3.0 * math.sqrt(5.0 / (2.0 * math.pi)) * _one_by_fivetwelve,
        0: 3.0 * math.sqrt(35.0 / math.pi) * _one_by_fivetwelve,
        +1: 3.0 * math.sqrt(5.0 / (2.0 * math.pi)) * _one_by_fivetwelve,
        +2: math.sqrt(15.0 / math.pi) * _one_by_fivetwelve,
        +3: -math.sqrt(15.0 / (2.0 * math.pi)) * _one_by_onetwentyeight,
        +4: math.sqrt(165.0 / (2.0 * math.pi)) * _one_by_sixteen,
        +5: -math.sqrt(165.0 / (2.0 * math.pi)) * _one_by_eight,
        +6: math.sqrt(2145.0 / math.pi) * 0.5,
        +7: -math.sqrt(15015.0 / (2.0 * math.pi)),
    },
    8: {
        -8: math.sqrt(34034.0 / math.pi),
        -7: math.sqrt(17017.0 / (2.0 * math.pi)),
        -6: math.sqrt(255255.0 / math.pi),
        -5: math.sqrt(12155.0 / (2.0 * math.pi)) * _one_by_eight,
        -4: math.sqrt(935.0 / (2.0 * math.pi)) * _one_by_thirtytwo,
        -3: math.sqrt(561.0 / (2.0 * math.pi)) * _one_by_onetwentyeight,
        -2: math.sqrt(17.0 / math.pi) * _one_by_fivetwelve,
        -1: math.sqrt(595.0 / (2.0 * math.pi)) * _one_by_fivetwelve,
        0: 3.0 * math.sqrt(595.0 / math.pi) * _one_by_fortyninetysix,
        +1: math.sqrt(595.0 / (2.0 * math.pi)) * _one_by_fivetwelve,
        +2: math.sqrt(17.0 / math.pi) * _one_by_fivetwelve,
        +3: -math.sqrt(561.0 / (2.0 * math.pi)) * _one_by_onetwentyeight,
        +4: math.sqrt(935.0 / (2.0 * math.pi)) * _one_by_thirtytwo,
        +5: -math.sqrt(12155.0 / (2.0 * math.pi)) * _one_by_eight,
        +6: math.sqrt(255255.0 / math.pi),
        +7: -math.sqrt(17017.0 / (2.0 * math.pi)),
        +8: math.sqrt(34034.0 / math.pi),
    },
}

# If cupy is loaded, put the coefficients into GPU memory.
# `_coeffs` dictionary will have entries for both numpy and cupy, depending on
# which is used.
# Also make a dictionary with the imaginary number in both numpy and cupy
# formats.
if cupy_here:
    _coeffs_cp = {
        l : {
            m: cupy.asarray(coeff)
            for m, coeff in six.iteritems(m_coeffs)
        }
        for l, m_coeffs in six.iteritems(_coeffs_np)
    }

    _coeffs = {
        np: _coeffs_np,
        cupy: _coeffs_cp,
    }
    _imag = {
        np: 1.0j,
        cupy: cupy.asarray(1.0j),
    }

# If cupy is not loaded, only entry will be for numpy.
else:
    _coeffs = {
        np: _coeffs_np,
    }
    _imag = {
        np: 1.0j,
    }



def SphericalHarmonicsVectorized(
        lm,
        theta, phi,
        xpy=xpy_default, dtype=np.complex128,
        l_max=8,
    ):
    """
    Compute spherical harmonics Y_{lm}(theta, phi) for a given set of lm's,
    thetas, and phis.

    Parameters
    ----------
    lm : array_like, shape = (n_indices, 2)
      Array of pairs of `l, m` values to evaluate harmonics for.
    theta : array_like, shape = (n_params,)
      Array of polar angles to evaluate harmonics at.
    phi : array_like, shape = (n_params,)
      Array of azimuth angles to evaluate harmonics at.
    xpy : numpy or cupy (default is cupy if loaded, else numpy)
      Numpy implementation to use.
    dtype : numpy.dtype (default is numpy.complex128)
      Datatype to use for output.  Must be complex.
    l_max : int (default is 8)
      Highest value of `l` that is allowed to appear in `lm`.  Will cause an
      exception if set lower than what's actually used.  If unsure, simply let
      it use the default, which uses the highest order the current
      implementation supports.

    Returns
    -------
    Ylm : array_like, shape = (n_params, n_indices)
      Array of spherical harmonics.  First axis varies `theta, phi`, and second
      axis varies `l, m`.
    """
    l, m = lm.T

    n_indices = l.size
    n_params = theta.size

    Ylm = xpy.empty((n_params, n_indices), dtype=dtype)

    coeffs = _coeffs[xpy]
    imag = _imag[xpy]

    ## Precompute commonly used functions of theta.
    # Values needed for all `l_max` values.
    cos_theta = xpy.cos(theta)
    sin_theta = xpy.sin(theta)

    one_minus_cos_theta = 1.0 - cos_theta
    one_plus_cos_theta = 1.0 + cos_theta

    # Values needed for `l_max` up to 3.
    if l_max >= 3:
        half_theta = 0.5 * theta
        two_theta = 2.0 * theta
        three_theta = 3.0 * theta

        cos_half_theta = xpy.cos(half_theta)
        sin_half_theta = xpy.sin(half_theta)

        sin_two_theta = xpy.sin(two_theta)
        sin_three_theta = xpy.sin(three_theta)

    # Values needed for `l_max` up to 4.
    if l_max >= 4:
        four_theta = 4.0 * theta

        cos_two_theta = xpy.cos(two_theta)
        sin_four_theta = xpy.sin(four_theta)

    # Values needed for `l_max` up to 5.
    if l_max >= 5:
        five_theta = 5.0 * theta

        cos_three_theta = xpy.cos(three_theta)
        sin_five_theta = xpy.sin(five_theta)

    # Values needed for `l_max` up to 6.
    if l_max >= 6:
        cos_four_theta = xpy.cos(four_theta)

    # Values needed for `l_max` up to 7.
    if l_max >= 7:
        cos_five_theta = xpy.cos(five_theta)

    # Values needed for `l_max` up to 8.
    if l_max >= 8:
        six_theta = 6.0 * theta

        cos_six_theta = xpy.cos(six_theta)


    for i, (l_i, m_i) in enumerate(lm):
        l_i, m_i = int(l_i), int(m_i)
        coeff = coeffs.get(l_i, {}).get(m_i, None)
        if l_i == 2:
            if m_i == -2:
                Ylm[...,i] = (
                    coeff *
                    xpy.square(one_minus_cos_theta)
                )
            elif m_i == -1:
                Ylm[...,i] = (
                    coeff *
                    sin_theta * one_minus_cos_theta
                )
            elif m_i == 0:
                Ylm[...,i] = (
                    coeff *
                    xpy.square(sin_theta)
                )
            elif m_i == +1:
                Ylm[...,i] = (
                    coeff *
                    sin_theta * one_plus_cos_theta
                )
            elif m_i == +2:
                Ylm[...,i] = (
                    coeff *
                    xpy.square(one_plus_cos_theta)
                )
        elif l_i == 3:
            if m_i == -3:
                Ylm[...,i] = (
                    coeff *
                    cos_half_theta * xpy.power(sin_half_theta, 5.0)
                )
            elif m_i == -2:
                Ylm[...,i] = (
                    coeff *
                    (2.0 + 3.0*cos_theta) * xpy.power(sin_half_theta, 4.0)
                )
            elif m_i == -1:
                Ylm[...,i] = (
                    coeff *
                    (sin_theta + 4.0*sin_two_theta - 3.0*sin_three_theta)
                )
            elif m_i == 0:
                Ylm[...,i] = (
                    coeff *
                    cos_theta * xpy.power(sin_theta, 2.0)
                )
            elif m_i == +1:
                Ylm[...,i] = (
                    coeff *
                    (sin_theta - 4.0*sin_two_theta - 3.0*sin_three_theta)
                )
            elif m_i == +2:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 4.0) *
                    (-2.0 + 3.0*cos_theta)
                )
            elif m_i == +3:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 5.0) *
                    sin_half_theta
                )
        elif l_i == 4:
            if m_i == -4:
                Ylm[...,i] = (
                    coeff *
                    xpy.square(cos_half_theta) *
                    xpy.power(sin_half_theta, 6.0)
                )
            elif m_i == -3:
                Ylm[...,i] = (
                    coeff *
                    cos_half_theta *
                    (1.0 + 2.0*cos_theta) *
                    xpy.power(sin_half_theta, 5.0)
                )
            elif m_i == -2:
                Ylm[...,i] = (
                    coeff *
                    (9.0 + 14.0*cos_theta + 7.0*cos_two_theta) *
                    xpy.power(sin_half_theta, 4.0)
                )
            elif m_i == -1:
                Ylm[...,i] = (
                    coeff *
                    (
                        3.0*sin_theta +
                        2.0*sin_two_theta +
                        7.0*sin_three_theta -
                        7.0*sin_four_theta
                    )
                )
            elif m_i == 0:
                Ylm[...,i] = (
                    coeff *
                    (5.0 + 7.0*cos_two_theta) *
                    xpy.square(sin_theta)
                )
            elif m_i == +1:
                Ylm[...,i] = (
                    coeff *
                    (
                        3.0*sin_theta -
                        2.0*sin_two_theta +
                        7.0*sin_three_theta +
                        7.0*sin_four_theta
                    )
                )
            elif m_i == +2:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 4.0) *
                    (9.0 - 14.0*cos_theta + 7.0*cos_two_theta)
                )
            elif m_i == +3:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 5.0) *
                    (-1.0 + 2.0*cos_theta) *
                    sin_half_theta
                )
            elif m_i == +4:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 6.0) *
                    xpy.square(sin_half_theta)
                )
        elif l_i == 5:
            if m_i == -5:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 3.0) *
                    xpy.power(sin_half_theta, 7.0)
                )
            elif m_i == -4:
                Ylm[...,i] = (
                    coeff *
                    xpy.square(cos_half_theta) *
                    (2.0 + 5.0*cos_theta) *
                    xpy.power(sin_half_theta, 6.0)
                )
            elif m_i == -3:
                Ylm[...,i] = (
                    coeff *
                    cos_half_theta *
                    (17.0 + 24.0*cos_theta + 15.0*cos_two_theta) *
                    xpy.power(sin_half_theta, 5.0)
                )
            elif m_i == -2:
                Ylm[...,i] = (
                    coeff *
                    (
                        32.0 +
                        57.0*cos_theta +
                        36.0*cos_two_theta +
                        15.0*cos_three_theta
                    ) *
                    xpy.power(sin_half_theta, 4.0)
                )
            elif m_i == -1:
                Ylm[...,i] = (
                    coeff *
                    (
                        2.0*sin_theta +
                        8.0*sin_two_theta +
                        3.0*sin_three_theta +
                        12.0*sin_four_theta -
                        15.0*sin_five_theta
                    )
                )
            elif m_i == 0:
                Ylm[...,i] = (
                    coeff *
                    (5.0*cos_theta + 3.0*cos_three_theta) *
                    xpy.square(sin_theta)
                )
            elif m_i == +1:
                Ylm[...,i] = (
                    coeff *
                    (
                        -2.0*sin_theta +
                        8.0*sin_two_theta -
                        3.0*sin_three_theta +
                        12.0*sin_four_theta +
                        15.0*sin_five_theta
                    )
                )
            elif m_i == +2:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 4.0) *
                    (
                        -32.0 +
                        57.0*cos_theta -
                        36.0*cos_two_theta +
                        15.0*cos_three_theta
                    )
                )
            elif m_i == +3:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 5.0) *
                    (17.0 - 24.0*cos_theta + 15.0*cos_two_theta) *
                    sin_half_theta
                )
            elif m_i == +4:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 6.0) *
                    (-2.0 + 5.0*cos_theta) *
                    xpy.square(sin_half_theta)
                )
            elif m_i == +5:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 7.0) *
                    xpy.power(sin_half_theta, 3.0)
                )
        elif l_i == 6:
            if m_i == -6:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 4.0) *
                    xpy.power(sin_half_theta, 8.0)
                )
            elif m_i == -5:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 3.0) *
                    (1.0 + 3.0*cos_theta) *
                    xpy.power(sin_half_theta, 7.0)
                )
            elif m_i == -4:
                Ylm[...,i] = (
                    coeff *
                    xpy.square(cos_half_theta) *
                    (35.0 + 44.0*cos_theta + 33.0*cos_two_theta) *
                    xpy.power(sin_half_theta, 6.0)
                )
            elif m_i == -3:
                Ylm[...,i] = (
                    coeff *
                    cos_half_theta *
                    (
                        98.0 +
                        185.0*cos_theta +
                        110.0*cos_two_theta +
                        55.0*cos_three_theta
                    ) *
                    xpy.power(sin_half_theta, 5.0)
                )
            elif m_i == -2:
                Ylm[...,i] = (
                    coeff *
                    (
                        1709.0 +
                        3096.0*cos_theta +
                        2340.0*cos_two_theta +
                        1320.0*cos_three_theta +
                        495.0*cos_four_theta
                    ) *
                    xpy.power(sin_half_theta, 4.0)
                )
            elif m_i == -1:
                Ylm[...,i] = (
                    coeff *
                    cos_half_theta *
                    (
                        161.0 +
                        252.0*cos_theta +
                        252.0*cos_two_theta +
                        132.0*cos_three_theta +
                        99.0*cos_four_theta
                    ) *
                    xpy.power(sin_half_theta, 3.0)
                )
            elif m_i == 0:
                Ylm[...,i] = (
                    coeff *
                    (35.0 + 60.0*cos_two_theta + 33.0*cos_four_theta) *
                    xpy.square(sin_theta)
                )
            elif m_i == +1:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 3.0) *
                    (
                        161.0 -
                        252.0*cos_theta +
                        252.0*cos_two_theta -
                        132.0*cos_three_theta +
                        99.0*cos_four_theta
                    ) *
                    sin_half_theta
                )
            elif m_i == +2:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 4.0) *
                    (
                        1709.0 -
                        3096.0*cos_theta +
                        2340.0*cos_two_theta -
                        1320.0*cos_three_theta +
                        495.0*cos_four_theta
                    )
                )
            elif m_i == +3:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 5.0) *
                    (
                        -98.0 +
                        185.0*cos_theta -
                        110.0*cos_two_theta +
                        55.0*cos_three_theta
                    ) *
                    sin_half_theta
                )
            elif m_i == +4:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 6.0) *
                    (35.0 - 44.0*cos_theta + 33.0*cos_two_theta) *
                    xpy.square(sin_half_theta)
                )
            elif m_i == +5:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 7.0) *
                    (-1.0 + 3.0*cos_theta) *
                    xpy.power(sin_half_theta, 3.0)
                )
            elif m_i == +6:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 8.0) *
                    xpy.power(sin_half_theta, 4.0)
                )
        elif l_i == 7:
            if m_i == -7:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 5.0) *
                    xpy.power(sin_half_theta, 9.0)
                )
            elif m_i == -6:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 4.0) *
                    (2.0 + 7.0*cos_theta) *
                    xpy.power(sin_half_theta, 8.0)
                )
            elif m_i == -5:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 3.0) *
                    (93.0 + 104.0*cos_theta + 91.0*cos_two_theta) *
                    xpy.power(sin_half_theta, 7.0)
                )
            elif m_i == -4:
                Ylm[...,i] = (
                    coeff *
                    xpy.square(cos_half_theta) *
                    (
                        140.0 +
                        285.0*cos_theta +
                        156.0*cos_two_theta +
                        91.0*cos_three_theta
                    ) *
                    xpy.power(sin_half_theta, 6.0)
                )
            elif m_i == -3:
                Ylm[...,i] = (
                    coeff *
                    cos_half_theta *
                    (
                        3115.0 +
                        5456.0*cos_theta +
                        4268.0*cos_two_theta +
                        2288.0*cos_three_theta +
                        1001.0*cos_four_theta
                    ) *
                    xpy.power(sin_half_theta, 5.0)
                )
            elif m_i == -2:
                Ylm[...,i] = (
                    coeff *
                    (
                        5220.0 +
                        9810.0*cos_theta +
                        7920.0*cos_two_theta +
                        5445.0*cos_three_theta +
                        2860.0*cos_four_theta +
                        1001.0*cos_five_theta
                    ) *
                    xpy.power(sin_half_theta, 4.0)
                )
            elif m_i == -1:
                Ylm[...,i] = (
                    coeff *
                    cos_half_theta *
                    (
                        1890.0 +
                        4130.0*cos_theta +
                        3080.0*cos_two_theta +
                        2805.0*cos_three_theta +
                        1430.0*cos_four_theta +
                        1001.0*cos_five_theta
                    ) *
                    xpy.power(sin_half_theta, 3.0)
                )
            elif m_i == 0:
                Ylm[...,i] = (
                    coeff *
                    cos_theta *
                    (109.0 + 132.0*cos_two_theta + 143.0*cos_four_theta) *
                    xpy.square(sin_theta)
                )
            elif m_i == +1:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 3.0) *
                    (
                        -1890.0 +
                        4130.0*cos_theta -
                        3080.0*cos_two_theta +
                        2805.0*cos_three_theta -
                        1430.0*cos_four_theta +
                        1001.0*cos_five_theta
                    ) *
                    sin_half_theta
                )
            elif m_i == +2:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 4.0) *
                    (
                        -5220.0 +
                        9810.0*cos_theta -
                        7920.0*cos_two_theta +
                        5445.0*cos_three_theta -
                        2860.0*cos_four_theta +
                        1001.0*cos_five_theta
                    )
                )
            elif m_i == +3:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 5.0) *
                    (
                        3115.0 -
                        5456.0*cos_theta +
                        4268.0*cos_two_theta -
                        2288.0*cos_three_theta +
                        1001.0*cos_four_theta
                    ) *
                    sin_half_theta
                )
            elif m_i == +4:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 6.0) *
                    (
                        -140.0 +
                        285.0*cos_theta -
                        156.0*cos_two_theta +
                        91.0*cos_three_theta
                    ) *
                    xpy.square(sin_half_theta)
                )
            elif m_i == +5:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 7.0) *
                    (93.0 - 104.0*cos_theta + 91.0*cos_two_theta) *
                    xpy.power(sin_half_theta, 3.0)
                )
            elif m_i == +6:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 8.0) *
                    (-2.0 + 7.0*cos_theta) *
                    xpy.power(sin_half_theta, 4.0)
                )
            elif m_i == +7:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 9.0) *
                    xpy.power(sin_half_theta, 5.0)
                )
        elif l_i == 8:
            if m_i == -8:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 6.0) *
                    xpy.power(sin_half_theta, 10.0)
                )
            elif m_i == -7:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 5.0) *
                    (1.0 + 4.0*cos_theta) *
                    xpy.power(sin_half_theta, 9.0)
                )
            elif m_i == -6:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 4.0) *
                    (1.0 + 2.0*cos_theta) *
                    xpy.sin(0.25*xpy.pi - half_theta) *
                    xpy.sin(0.25*xpy.pi + half_theta) *
                    xpy.power(sin_half_theta, 8.0)
                )
            elif m_i == -5:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 3.0) *
                    (
                        19.0 +
                        42.0*cos_theta +
                        21.0*cos_two_theta +
                        14.0*cos_three_theta
                    ) *
                    xpy.power(sin_half_theta, 7.0)
                )
            elif m_i == -4:
                Ylm[...,i] = (
                    coeff *
                    xpy.square(cos_half_theta) *
                    (
                        265.0 +
                        442.0*cos_theta +
                        364.0*cos_two_theta +
                        182.0*cos_three_theta +
                        91.0*cos_four_theta
                    ) *
                    xpy.power(sin_half_theta, 6.0)
                )
            elif m_i == -3:
                Ylm[...,i] = (
                    coeff *
                    cos_half_theta *
                    (
                        869.0 +
                        1660.0*cos_theta +
                        1300.0*cos_two_theta +
                        910.0*cos_three_theta +
                        455.0*cos_four_theta +
                        182.0*cos_five_theta
                    ) *
                    xpy.power(sin_half_theta, 5.0)
                )
            elif m_i == -2:
                Ylm[...,i] = (
                    coeff *
                    (
                        7626.0 +
                        14454.0*cos_theta +
                        12375.0*cos_two_theta +
                        9295.0*cos_three_theta +
                        6006.0*cos_four_theta +
                        3003.0*cos_five_theta +
                        1001.0*cos_six_theta
                    ) *
                    xpy.power(sin_half_theta, 4.0)
                )
            elif m_i == -1:
                Ylm[...,i] = (
                    coeff *
                    cos_half_theta *
                    (
                        798.0 +
                        1386.0*cos_theta +
                        1386.0*cos_two_theta +
                        1001.0*cos_three_theta +
                        858.0*cos_four_theta +
                        429.0*cos_five_theta +
                        286.0*cos_six_theta
                    ) *
                    xpy.power(sin_half_theta, 3.0)
                )
            elif m_i == 0:
                Ylm[...,i] = (
                    coeff *
                    (
                        210.0 +
                        385.0*cos_two_theta +
                        286.0*cos_four_theta +
                        143.0*cos_six_theta
                    ) *
                    xpy.square(sin_theta)
                )
            elif m_i == +1:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 3.0) *
                    (
                        798.0 -
                        1386.0*cos_theta +
                        1386.0*cos_two_theta -
                        1001.0*cos_three_theta +
                        858.0*cos_four_theta -
                        429.0*cos_five_theta +
                        286.0*cos_six_theta
                    ) *
                    sin_half_theta
                )
            elif m_i == +2:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 4.0) *
                    (
                        7626.0 -
                        14454.0*cos_theta +
                        12375.0*cos_two_theta -
                        9295.0*cos_three_theta +
                        6006.0*cos_four_theta -
                        3003.0*cos_five_theta +
                        1001.0*cos_six_theta
                    )
                )
            elif m_i == +3:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 5.0) *
                    (
                        -869.0 +
                        1660.0*cos_theta -
                        1300.0*cos_two_theta +
                        910.0*cos_three_theta -
                        455.0*cos_four_theta +
                        182.0*cos_five_theta
                    ) *
                    sin_half_theta
                )
            elif m_i == +4:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 6.0) *
                    (
                        265.0 -
                        442.0*cos_theta +
                        364.0*cos_two_theta -
                        182.0*cos_three_theta +
                        91.0*cos_four_theta
                    ) *
                    xpy.square(sin_half_theta)
                )
            elif m_i == +5:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 7.0) *
                    (
                        -19.0 +
                        42.0*cos_theta -
                        21.0*cos_two_theta +
                        14.0*cos_three_theta
                    ) *
                    xpy.power(sin_half_theta, 3.0)
                )
            elif m_i == +6:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 8.0) *
                    (-1.0 + 2.0*cos_theta) *
                    xpy.sin(0.25*xpy.pi - half_theta) *
                    xpy.sin(0.25*xpy.pi + half_theta) *
                    xpy.power(sin_half_theta, 4.0)
                )
            elif m_i == +7:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 9.0) *
                    (-1.0 + 4.0*cos_theta) *
                    xpy.power(sin_half_theta, 5.0)
                )
            elif m_i == +8:
                Ylm[...,i] = (
                    coeff *
                    xpy.power(cos_half_theta, 10.0) *
                    xpy.power(sin_half_theta, 6.0)
                )


        # Chosen l, m values either not implemented yet or unphysical.
        else:
            Ylm[...,i] = xpy.nan
    # Multiply everything by exp(i * m * phi)
    Ylm *= xpy.exp(imag * xpy.einsum("i,j", phi, m))

    return Ylm



#! /usr/bin/env python
#  
#
# Tool: dumps a candidate waveform to a frame file.
# Default GPS time used is boring
# Tries to be as close to C as possible -- no interface via pylal/glue
#
# EXAMPLE
#    python util_LALWriteFrame.py; FrDump -i dummy.gwf
#
#  WARNING: My version does NOT interpolate the signal to a synchronized set of sample times.
#                       This may cause problems for some applications, particularly at low sample rates.


import argparse
import numpy as np
import RIFT.lalsimutils as lalsimutils
import lalsimulation as lalsim
import lalframe
import lal

import RIFT.physics.ROMWaveformManager as romwf

parser = argparse.ArgumentParser()
parser.add_argument("--group", default="my_surrogates/nr_surrogates/",help="Surrogate example tutorial/TutorialSurrogate")
parser.add_argument("--param", default='/SpEC_q1_9_NoSpin_SingleModes_REF.h5',help="Parameter value:  /SpEC_q1_10_NoSpin_nu5thDegPoly_exclude_2_0_FastSplined.h5, ...")
parser.add_argument("--lmax", type=int, default=2,help="Set max modes")
parser.add_argument("--fname", default=None, help = "Base name for output frame file. Otherwise auto-generated ")
parser.add_argument("--instrument", default="H1",help="Use H1, L1,V1")
parser.add_argument("--inj", dest='inj', default=None,help="inspiral XML file containing injection information.")
parser.add_argument("--event",type=int, dest="event_id", default=None,help="event ID of injection XML to use.")
parser.add_argument("--single-ifo",default=False,action='store_true')
parser.add_argument("--approx",type=str,default=None)
parser.add_argument("--srate",type=int,default=16384,help="Sampling rate")
parser.add_argument("--seglen", type=float,default=16., help="Default window size for processing.")
parser.add_argument("--start", type=int,default=None)
parser.add_argument("--stop", type=int,default=None)
parser.add_argument("--fref", dest='fref', type=float, default=0.0, help="Waveform reference frequency [template]. Required, default is 0 (coalescence).")
parser.add_argument("--incl",default=None,help="Set the inclination of L (at fref). Particularly helpful for aligned spin tests")
parser.add_argument("--mass1",default=35,type=float,help='Mass 1 (solar masses)')
parser.add_argument("--mass2",default=34,type=float,help='Mass 2 (solar masses)')
parser.add_argument("--verbose", action="store_true",default=False)
opts=  parser.parse_args()


# Generate signal
P = lalsimutils.ChooseWaveformParams()
P.deltaT = 1./opts.srate
P.radec = True  # use a real source with a real instrument
if not opts.inj:
    P.randomize(aligned_spin_Q=True,default_inclination=opts.incl)
    P.m1 = opts.mass1*lalsimutils.lsu_MSUN
    P.m2 = opts.mass2*lalsimutils.lsu_MSUN
    P.taper = lalsimutils.lsu_TAPER_START
    P.tref =1000000000  # default
    if opts.approx:
        P.approx = lalsim.GetApproximantFromString(str(opts.approx))
    else:
        P.approx = lalsim.GetApproximantFromString("SpinTaylorT2")
else:
    from glue.ligolw import lsctables, table, utils # check all are needed

    filename = opts.inj
    event = opts.event_id
    xmldoc = utils.load_filename(filename, verbose = True, contenthandler =lalsimutils.cthdler)
    sim_inspiral_table = table.get_table(xmldoc, lsctables.SimInspiralTable.tableName)
    P.copy_sim_inspiral(sim_inspiral_table[int(event)])
    if opts.approx:
        P.approx=lalsimutils.StringToLALApproximant(opts.approx)
P.taper = lalsimutils.lsu_TAPER_START
P.detector = opts.instrument
P.print_params()


T_est = lalsimutils.estimateWaveformDuration(P)
T_est = P.deltaT*lalsimutils.nextPow2(T_est/P.deltaT)
T_est = np.max([4,T_est])
P.deltaF = 1./T_est
print " Duration ", T_est
if T_est < opts.seglen:
    print " Buffer length too short, automating retuning forced "

# Generate ROM
acatHere = romwf.WaveformModeCatalog(opts.group, opts.param, opts.lmax)

# Generate signal
hoft = acatHere.real_hoft(P)   # include translation of source, but NOT interpolation onto regular time grid
# zero pad to be opts.seglen long, if necessary
if opts.seglen/hoft.deltaT > hoft.data.length:
    TDlenGoal = int(opts.seglen/hoft.deltaT)
    hoft = lal.ResizeREAL8TimeSeries(hoft, 0, TDlenGoal)

# zero pad some more on either side, to make sure the segment covers start to stop
if opts.start and hoft.epoch > opts.start:
    nToAddBefore = int((hoft.epoch-opts.start)/hoft.deltaT)
    print nToAddBefore, hoft.data.length
    ht = lal.CreateREAL8TimeSeries("Template h(t)", 
            hoft.epoch - nToAddBefore*hoft.deltaT, 0, hoft.deltaT, lalsimutils.lsu_DimensionlessUnit, 
            hoft.data.length+nToAddBefore)
    ht.data.data = np.zeros(ht.data.length)  # clear
    ht.data.data[nToAddBefore:nToAddBefore+hoft.data.length] = hoft.data.data
    hoft = ht

if opts.stop and hoft.epoch+hoft.data.length*hoft.deltaT < opts.stop:
    nToAddAtEnd = int( (-(hoft.epoch+hoft.data.length*hoft.deltaT)+opts.stop)/hoft.deltaT)
    print "Padding end ", nToAddAtEnd, hoft.data.length
    hoft = lal.ResizeREAL8TimeSeries(hoft,0, int(hoft.data.length+nToAddAtEnd))

channel = opts.instrument+":FAKE-STRAIN"

tstart = int(hoft.epoch)
duration = int(hoft.data.length*hoft.deltaT)
if not opts.fname:
    fname = opts.instrument.replace("1","")+"-fake_strain-"+str(tstart)+"-"+str(duration)+".gwf"

print "Writing signal with ", hoft.data.length*hoft.deltaT, " to file ", fname
lalsimutils.hoft_to_frame_data(fname,channel,hoft)

# TEST: Confirm it works by reading the frame
if opts.verbose:
    import os
    from matplotlib import pyplot as plt
    plt.switch_backend('agg')  # fix backend
    # First must create corresponding cache file
    os.system("echo "+ fname+ " | lalapps_path2cache   > test.cache")
    # Now I can read it
    # Beware that the results are OFFSET FROM ONE ANOTHER due to PADDING,
    #    but that the time associations are correct
    hoft2 = lalsimutils.frame_data_to_hoft("test.cache", channel)
    tvals2 = (float(hoft2.epoch) - float(P.tref)) +  np.arange(hoft2.data.length)*hoft2.deltaT
    tvals = (float(hoft.epoch) - float(P.tref)) +  np.arange(hoft.data.length)*hoft.deltaT
    plt.plot(tvals2,hoft2.data.data,label='Fr')
    plt.plot(tvals,hoft.data.data,label='orig')
    plt.xlim(-1,0.1)  # should scale with mass
    plt.legend(); 
    plt.savefig("frdump_rom_"+opts.instrument+".png")



#! /usr/bin/env python
#
#
# GOAL
#   - takes two sets of samples, and some parameter(s)
#       - should be able to interchange samples provided with ILE *.xml.gz, *.composite, or posterior samples (preferred).  FLEXIBILITY NOT YET IMPLEMENTED. 
#         Postfix determines behavior
#   - performs specified test, with specified tolerance, to see if they are 'similar enough'
#   - returns FAILURE if test is a success (!), so a condor DAG will terminate
#
# EXAMPLES
# convergence_test_samples.py --samples GW170823_pure_NR_and_NRSur7dq2_lmax3_fmin20_C02_cleaned_alignedspin_zprior.dat --samples GW170823_pure_NR_and_NRSur7dq2_lmax3_fmin20_C02_cleaned_alignedspin_zprior.dat --parameter m1 --parameter m2   # test samples against themselves, must return 0!
#
# RESOURCES
#   Based on code in util_DriverIterateILEFitPosterior*.py

import numpy as np
import argparse
import scipy.stats
import numpy.linalg as la
import sys

from RIFT.misc.samples_utils import add_field

parser = argparse.ArgumentParser()
parser.add_argument("--samples", action='append', help="Samples used in convergence test")
parser.add_argument("--parameter", action='append', help="Parameters used in convergence test")
parser.add_argument("--parameter-range", action='append', help="Parameter ranges used in convergence test (used if KDEs or similar knowledge of the PDF is needed). If used, must specify for ALL variables, in order")
parser.add_argument("--method", default='lame',  help="Test to perform: lame|ks1d|...")
parser.add_argument("--threshold",default=0.01,type=float,  help="Manual threshold for the test being performed. (If not specified, the success condition is determined by default for that diagnostic, based on the samples size and properties).  Try 0.01")
parser.add_argument("--test-output",  help="Filename to return output. Result is a scalar >=0 and ideally <=1.  Closer to 0 should be good. Second column is the diagnostic, first column is 0 or 1 (success or failure)")
parser.add_argument("--always-succeed",action='store_true',help="Test output is always success.  Use for plotting convergence diagnostics so jobs insured to run for many iterations.")
opts=  parser.parse_args()

if len(opts.samples)<1:
    print " Need at least two sets of samples"
    sys.exit(0)




# Test options
#
#   (a) lame: Compute a multivariate gaussian estimate (sample mean and variance), and then use KL divergence between them !
#   (b) KS_1d: One-dimensional KS test on cumulative distribution  
#   (c) KL_1d: One-dimensional KL divergence, using KDE estimate.  Requires bounded domain; parameter bounds can be passed 


def calc_kl(mu_1, mu_2, sigma_1, sigma_2, sigma_1_inv, sigma_2_inv):
    """
    calc_kl : KL divergence for two gaussians.  sigma_1, and sigma_2 are the covariance matricies.
    """
    return 0.5*(np.trace(np.dot(sigma_2_inv,sigma_1))+np.dot(np.dot((mu_2-mu_1).T, sigma_2_inv), (mu_2-mu_1))-len(mu_1)+np.log(la.det(sigma_2)/la.det(sigma_1)))

def calc_kl_scalar(mu_1, mu_2, sigma_1, sigma_2):
    """
    calc_kl : KL divergence for two gaussians.  sigma_1, and sigma_2 are the covariance matricies.
    """
    return np.log(sigma_2/sigma_1) -0.5 +( (mu_1-mu_2)**2 + sigma_1**2)/(2*sigma_2**2)


def test_lame(dat1,dat2):
    """
    Compute a multivariate gaussian estimate (sample mean and variance), and then use KL divergence between them !
    """
    mu_1 = np.mean(dat1,axis=0)
    mu_2 = np.mean(dat2,axis=0)
    sigma_1 = np.cov(dat1.T)
    sigma_2 = np.cov(dat2.T)
    if np.isscalar(mu_1) or len(mu_1)==1:
        return np.asscalar(calc_kl_scalar(mu_1, mu_2, sigma_1, sigma_2))
    else:
        sigma_1_inv = np.linalg.inv(sigma_1)
        sigma_2_inv = np.linalg.inv(sigma_2)
    return calc_kl(mu_1,mu_2, sigma_1, sigma_2, sigma_1_inv, sigma_2_inv)

def test_ks1d(dat1_1d, dat2_1d):
    """
    KS test based on two sample sets.  Uses the KS D value as threshold
    https://docs.scipy.org/doc/scipy-0.14.0/reference/generated/scipy.stats.ks_2samp.html
    """
    return scipy.stats.ks_2samp(dat1_1d,dat2_1d)[0]  # return KS statistic

def test_KL1d(dat1_1d,dat2_1d,range1=None, range2=None):
    return None




# Procedure

samples1 = np.genfromtxt(opts.samples[0], names=True)
samples2 = np.genfromtxt(opts.samples[1], names=True)

# Add missing fields needed for some tests
if not('xi' in samples1.dtype.names):
            samples1 = add_field(samples1, [('chi_eff',float)]); samples1['chi_eff'] = (samples1["m1"]*samples1["a1z"]+samples1["m2"]*samples1["a2z"])/(samples1["m1"]+samples1["m2"])
            samples1 = add_field(samples1, [('xi',float)]); samples1['xi'] = (samples1["m1"]*samples1["a1z"]+samples1["m2"]*samples1["a2z"])/(samples1["m1"]+samples1["m2"])
if not('xi' in samples2.dtype.names):
            samples2 = add_field(samples2, [('chi_eff',float)]); samples2['chi_eff'] = (samples2["m1"]*samples2["a1z"]+samples2["m2"]*samples2["a2z"])/(samples2["m1"]+samples2["m2"])
            samples2 = add_field(samples2, [('xi',float)]); samples2['xi'] = (samples2["m1"]*samples2["a1z"]+samples2["m2"]*samples2["a2z"])/(samples2["m1"]+samples2["m2"])


param_names1 = samples1.dtype.names; param_names2 = samples2.dtype.names
npts1 = len(samples1[param_names1[0]])
npts2 = len(samples2[param_names2[0]])  

# Read in data into array.  For now, assume the specific parameters requested are provided.
dat1 = np.empty( (npts1,len(opts.parameter)))
dat2 = np.empty( (npts2,len(opts.parameter)))
indx=0
for param in opts.parameter:
    dat1[:,indx] = samples1[param]
    dat2[:,indx] = samples2[param]
    indx+=1


# Perform test
val_test = np.inf
if opts.method == 'lame':
    val_test = test_lame(dat1,dat2)
elif opts.method == 'KS_1d':
    val_test = test_ks1d(dat1[:,0],dat2[:,0])
elif opts.method == 'KL_1d':
    val_test = test_KL1d(dat1[:,0],dat2[:,0])
else:
    print " No known method ", opts.method
print  val_test

if opts.always_succeed or (opts.threshold is None):
    sys.exit(0)

if (val_test < opts.threshold):  
    sys.exit(1)
else:
    sys.exit(0)



import numpy
try:
    import cupy
    xpy_default=cupy
    junk_to_check_installed = cupy.array(5)  # this will fail if GPU not installed correctly
except:
    xpy_default=numpy



def TimeDelayFromEarthCenter(
        detector_earthfixed_xyz_metres,
        source_right_ascension_radians,
        source_declination_radians,
        greenwich_mean_sidereal_time,
        xpy=xpy_default, dtype=numpy.float64,
    ):
    """

    Parameters
    ----------
    detector_earthfixed_xyz_metres : array_like, shape = det_shape + (3,)
      Location of detector(s) relative to Earth's center in meters. May provide
      multiple detectors, last axis must be (x,y,z) but other axes can take
      whatever form is desired.
    source_right_ascension_radians : array_like, shape = sample_shape
      Right ascension of source in radians, can be an arbitrary dimensional
      array.
    source_declination_radians : array_like, shape = sample_shape
      Declination of source in radians, can be an arbitrary dimensional array.
    greenwich_mean_sidereal_time : float
      Should be equivalent to XLALGreenwichMeanSiderealTime(gpstime).

    Returns
    -------
    time_delay_from_earth_center : array_like, shape = det_shape + sample_shape
    """
    negative_speed_of_light = xpy.asarray(-299792458.0)

    det_shape = detector_earthfixed_xyz_metres.shape[:-1]
    sample_shape = source_right_ascension_radians.shape

    cos_dec = xpy.cos(source_declination_radians)

    greenwich_hour_angle = (
        greenwich_mean_sidereal_time - source_right_ascension_radians
    )

    ehat_src = xpy.empty(sample_shape + (3,), dtype=dtype)

    ehat_src[...,0] = cos_dec * xpy.cos(greenwich_hour_angle)
    ehat_src[...,1] = -cos_dec * xpy.sin(greenwich_hour_angle)
    ehat_src[...,2] = xpy.sin(source_declination_radians)

    neg_separation = xpy.inner(detector_earthfixed_xyz_metres, ehat_src)
    return xpy.divide(
        neg_separation, negative_speed_of_light,
        out=neg_separation,
    )


def ComputeDetAMResponse(
        detector_response_matrix,
        source_right_ascension_radians,
        source_declination_radians,
        source_polarization_radians,
        greenwich_mean_sidereal_time,
        xpy=xpy_default, dtype_real=numpy.float64, dtype_complex=numpy.complex128,
    ):
    """
    Parameters
    ----------
    detector_response_matrix : array_like, shape = det_shape + (3, 3)
      Detector response matrix, or matrices for multiple detectors.  Last two
      axes must be 3-by-3 response matrix, and may include arbitrary axes before
      that for various detectors.
    source_right_ascension_radians : array_like, shape = sample_shape
      Right ascension of source in radians, can be an arbitrary dimensional
      array.
    source_declination_radians : array_like, shape = sample_shape
      Declination of source in radians, can be an arbitrary dimensional array.
    source_polarization_radians : array_like, shape = sample_shape
      Polarization angle of source in radians, can be an arbitrary dimensional
      array.
    greenwich_mean_sidereal_time : float
      Should be equivalent to XLALGreenwichMeanSiderealTime(gpstime).

    Returns
    -------
    F : array_like, shape = det_shape + sample_shape
    """
    det_shape = detector_response_matrix.shape[:-1]
    sample_shape = source_right_ascension_radians.shape
    matrix_shape = 3, 3

    # Initialize trig matrices.
    X = xpy.empty(sample_shape+(3,), dtype=dtype_real)
    Y = xpy.empty(sample_shape+(3,), dtype=dtype_real)

    # Greenwich hour angle of source in radians.
    source_greenwich_radians = (
        greenwich_mean_sidereal_time - source_right_ascension_radians
    )

    # Pre-compute trig functions
    cos_gha = xpy.cos(source_greenwich_radians)
    sin_gha = xpy.sin(source_greenwich_radians)
    cos_dec = xpy.cos(source_declination_radians)
    sin_dec = xpy.sin(source_declination_radians)
    cos_psi = xpy.cos(source_polarization_radians)
    sin_psi = xpy.sin(source_polarization_radians)

    # Populate trig matrices.
    X[...,0] = -cos_psi*sin_gha - sin_psi*cos_gha*sin_dec
    X[...,1] = -cos_psi*cos_gha + sin_psi*sin_gha*sin_dec
    X[...,2] =  sin_psi*cos_dec

    Y[...,0] =  sin_psi*sin_gha - cos_psi*cos_gha*sin_dec
    Y[...,1] =  sin_psi*cos_gha + cos_psi*sin_gha*sin_dec
    Y[...,2] =  cos_psi*cos_dec

    # Compute F for each polarization state.
    F_plus = (
        X*xpy.inner(X, detector_response_matrix) -
        Y*xpy.inner(Y, detector_response_matrix)
    ).sum(axis=-1)
    F_cross = (
        X*xpy.inner(Y, detector_response_matrix) +
        Y*xpy.inner(X, detector_response_matrix)
    ).sum(axis=-1)

    return F_plus + 1.0j*F_cross



#! /usr/bin/env python
#
# GOAL
#
#
# HISTORY
#   - Based on testing_archival_and_pseudo_online/scripts/setup_analysis_gracedb_event.py  in  richard-oshaughnessy/rapid_pe_nr_review_o3.git
#
# EXAMPLES
#    Here, <EXE> refers to the name given to this code
#  - Reproduce argument sequence of lalinference_pipe
#       <EXE>  --use-ini `pwd`/test.ini --use-coinc `pwd`/coinc.xml --use-rundir `pwd`/test --use-online-psd-file `pwd`/psd.xml.gz
#  - Run on events with full automation 
#       <EXE> --gracedb-id G329483 --approx NRHybSur3dq8 --l-max 4


import numpy as np
import argparse
import os
import sys
import lal
import lalsimulation as lalsim
import RIFT.lalsimutils as lalsimutils

import shutil

# Default setup assumes the underlying sampling will be *cartesian* 
# for a precessing binary.  Change as appropriate if the underlying helper changes to be more sensible.
prior_args_lookup={}
prior_args_lookup["default"] =""
prior_args_lookup["volumetric"] =""
prior_args_lookup["uniform_mag_prec"] =" --pseudo-uniform-magnitude-prior "
prior_args_lookup["uniform_aligned"] =""
prior_args_lookup["zprior_aligned"] =" --aligned-prior alignedspin-zprior"

typical_bns_range_Mpc = {}
typical_bns_range_Mpc["O1"] = 100 
typical_bns_range_Mpc["O2"] = 100 
typical_bns_range_Mpc["O3"] = 130
observing_run_time ={}
observing_run_time["O1"] = [1126051217,1137254417] # https://www.gw-openscience.org/O1/
observing_run_time["O2"] = [1164556817,1187733618] # https://www.gw-openscience.org/O2/
observing_run_time["O3"] = [1230000000,1430000000] # Completely made up boundaries, for now
def get_observing_run(t):
    for run in observing_run_time:
        if  t > observing_run_time[run][0] and t < observing_run_time[run][1]:
            return run
    print " No run available for time ", t, " in ", observing_run_time
    return None

def unsafe_config_get(config,args,verbose=False):
    if verbose:
        print " Retrieving ", args, 
        print " Found ",eval(config.get(*args))
    return eval( config.get(*args))


def retrieve_event_from_coinc(fname_coinc):
    from glue.ligolw import lsctables, table, utils
    from RIFT import lalsimutils
    event_dict ={}
    samples = table.get_table(utils.load_filename(fname_coinc,contenthandler=lalsimutils.cthdler), lsctables.SnglInspiralTable.tableName)
    event_duration=4  # default
    ifo_list = []
    for row in samples:
        m1 = row.mass1
        m2 = row.mass2
        ifo_list.append(row.ifo)
        try:
            event_duration = row.event_duration # may not exist
        except:
            print " event_duration field not in XML "
    event_dict["m1"] = row.mass1
    event_dict["m2"] = row.mass2
    event_dict["s1z"] = row.spin1z
    event_dict["s2z"] = row.spin2z
    event_dict["tref"] = row.end_time + 1e-9*row.end_time_ns
    event_dict["IFOs"] = list(set(ifo_list))
    return event_dict


parser = argparse.ArgumentParser()
parser.add_argument("--use-ini",default=None,type=str,help="Pass ini file for parsing. Intended to reproduce lalinference_pipe functionality. Overrides most other arguments. Full path recommended")
parser.add_argument("--use-rundir",default=None,type=str,help="Intended to reproduce lalinference_pipe functionality. Must be absolute path.")
parser.add_argument("--use-online-psd-file",default=None,type=str,help="Provides specific online PSD file, so no downloads are needed")
parser.add_argument("--use-coinc",default=None,type=str,help="Intended to reproduce lalinference_pipe functionality")
parser.add_argument("--online",action='store_true')
parser.add_argument("--extra-args-helper",action=None, help="Filename with arguments for the helper. Use to provide alternative channel names and other advanced configuration (--channel-name, data type)!")
parser.add_argument("--manual-postfix",default='',type=str)
parser.add_argument("--gracedb-id",default=None,type=str)
parser.add_argument("--event-time",default=None,type=float,help="Event time. Intended to override use of GracedbID. MUST provide --manual-initial-grid ")
parser.add_argument("--calibration",default="C00",type=str)
parser.add_argument("--playground-data",action='store_true', help="Passed through to helper_LDG_events, and changes name prefix")
parser.add_argument("--approx",default=None,type=str,help="Approximant. REQUIRED")
parser.add_argument("--use-gwsurrogate",action='store_true',help="Attempt to use gwsurrogate instead of lalsuite.")
parser.add_argument("--l-max",default=2,type=int)
parser.add_argument("--no-matter",action='store_true', help="Force analysis without matter. Really only matters for BNS")
parser.add_argument("--assume-matter",action='store_true', help="Force analysis *with* matter. Really only matters for BNS")
parser.add_argument("--assume-highq",action='store_true', help="Force analysis with the high-q strategy, neglecting spin2. Passed to 'helper'")
parser.add_argument("--add-extrinsic",action='store_true')
parser.add_argument("--fmin",default=20,type=int,help="Mininum frequency for integration. template minimum frequency (we hope) so all modes resolved at this frequency")  # should be 23 for the BNS
parser.add_argument("--fmin-template",default=None,type=float,help="Mininum frequency for template. If provided, then overrides automated settings for fmin-template = fmin/Lmax")  # should be 23 for the BNS
parser.add_argument("--data-LI-seglen",default=None,type=int,help="If specified, passed to the helper. Uses data selection appropriate to LI. Must specify the specific LI seglen used.")
parser.add_argument("--choose-data-LI-seglen",action='store_true')
parser.add_argument("--fix-bns-sky",action='store_true')
parser.add_argument("--cip-sampler-method",type=str,default=None)
parser.add_argument("--cip-fit-method",type=str,default=None)
parser.add_argument("--spin-magnitude-prior",default='default',type=str,help="options are default [volumetric for precessing,uniform for aligned], volumetric, uniform_mag_prec, uniform_mag_aligned, zprior_aligned")
parser.add_argument("--force-chi-max",default=None,type=float,help="Provde this value to override the value of chi-max provided") 
parser.add_argument("--hierarchical-merger-prior-1g",action='store_true',help="As in 1903.06742")
parser.add_argument("--hierarchical-merger-prior-2g",action='store_true',help="As in 1903.06742")
parser.add_argument("--link-reference-pe",action='store_true',help="If present, creates a directory 'reference_pe' and adds symbolic links to fiducial samples. These can be used by the automated plotting code.  Requires LVC_PE_SAMPLES environment variable defined!")
parser.add_argument("--link-reference-psds",action='store_true',help="If present, uses the varialbe LVC_PE_CONFIG to find a 'reference_pe_config_map.dat' file, which provides the location for reference PSDs.  Will override PSDs used / setup by default")
parser.add_argument("--make-bw-psds",action='store_true',help='If present, adds nodes to create BW PSDs to the dag.  If at all possible, avoid this and re-use existing PSDs')
parser.add_argument("--link-bw-psds",action='store_true',help='If present, uses the script retrieve_bw_psd_for_event.sh  to find a precomputed BW psd, and convert it to our format')
parser.add_argument("--use-online-psd",action='store_true', help="If present, will use the online PSD estimates")
parser.add_argument("--ile-retries",default=3,type=int)
parser.add_argument("--fit-save-gp",action="store_true",help="If true, pass this argument to CIP. GP plot for each iteration will be saved. Useful for followup investigations or reweighting. Warning: lots of disk space (1G or so per iteration)")
parser.add_argument("--cip-explode-jobs",type=int,default=None)
parser.add_argument("--cip-quadratic-first",action='store_true')
parser.add_argument("--manual-initial-grid",default=None,type=str,help="Filename (full path) to initial grid. Copied into proposed-grid.xml.gz, overwriting any grid assignment done here")
parser.add_argument("--manual-extra-ile-args",default=None,type=str,help="Avenue to adjoin extra ILE arguments.  Needed for unusual configurations (e.g., if channel names are not being selected, etc)")
parser.add_argument("--verbose",action='store_true')
parser.add_argument("--use-osg",action='store_true',help="Restructuring for ILE on OSG. The code will TRY to make a copy of the necessary frame files, from the reference directory")
opts=  parser.parse_args()

if (opts.approx is None) and not (opts.use_ini is None):
    import ConfigParser
    config = ConfigParser.ConfigParser()
    config.read(opts.use_ini)
    approx_name_ini = config.get('engine','approx')
    approx_name_cleaned = lalsim.GetStringFromApproximant(lalsim.GetApproximantFromString(approx_name_ini))
    opts.approx = approx_name_cleaned
    print " Approximant provided in ini file: ",approx_name_cleaned
elif opts.approx is None:
    print " Approximant required! "
    sys.exit(1)

if opts.use_osg:
    os.environ["LIGO_DATAFIND_SERVER"]="datafind.ligo.org:443"   #  enable lookup of data

if opts.make_bw_psds:
    if not(opts.choose_data_LI_seglen) and (opts.data_LI_seglen is None):
        print " To use the BW PSD, you MUST provide a default analysis seglen "
        sys.exit(1)

if opts.online:
        opts.use_online_psd =True
        if opts.link_bw_psds:
            print " Inconsistent options for PSDs "
            sys.exit(1)

fmin = opts.fmin
fmin_template  = opts.fmin
if opts.l_max > 2:
    print " ==> Reducing minimum template frequency because of HM <== "
    fmin_template = opts.fmin * 2./opts.l_max
if not(opts.fmin_template is None):
    fmin_template = opts.fmin_template
gwid = opts.gracedb_id if (not opts.gracedb_id is None) else '';
if opts.gracedb_id is None:
    gwid="manual_"+ str(opts.event_time)
    if not (opts.use_ini is None):
        gwid = ''
else:
# checks X509_USER_PROXY env variable
# if empty, checks grid-proxy-info -path
# if empty, fails and tells you to run ligo-proxy-init
    if not("X509_USER_PROXY" in os.environ.keys()):
        import subprocess
        str_proxy =subprocess.check_output(['grid-proxy-info','-path']).rstrip()
        if len(str_proxy) < 1:
            print " Run ligo-proxy-init  or otherwise have a method to query gracedb / use CVMFS frames as you need! "
            sys.exit(1)
print " Event ", gwid
base_dir = os.getcwd()
if opts.use_ini:
    base_dir =''  # all directories are provided as full path names


if opts.choose_data_LI_seglen:
    cmd_event = "gracedb_legacy download " + opts.gracedb_id + " coinc.xml"
    os.system(cmd_event)
    event_dict = retrieve_event_from_coinc("coinc.xml")
    P=lalsimutils.ChooseWaveformParams()
    P.m1 = event_dict["m1"]*lal.MSUN_SI; P.m2=event_dict["m2"]*lal.MSUN_SI; P.s1z = event_dict["s1z"]; P.s2z = event_dict["s2z"]
    P.fmin = opts.fmin  #  fmin we will use internally
    T_wave = lalsimutils.estimateWaveformDuration(P) +2  # 2 second buffer on end; note that with next power of 2, will go up to 4s
    T_wave_round = lalsimutils.nextPow2( T_wave)

    # For frequency-domain approximants, I need another factor of 2!
    # We have an extra buffer
    if lalsim.SimInspiralImplementedFDApproximants(P.approx)==1:
            print " FD approximant, needs extra buffer for RIFT at present "
            T_wave_round *=2 

    print " Assigning auto-selected segment length ", T_wave_round
    opts.data_LI_seglen  = T_wave_round

    # Problem with SEOBNRv3 starting frequencies
    mtot_msun = event_dict["m1"]+event_dict["m2"] 
    if ('SEOB' in opts.approx) and mtot_msun > 90*(20./opts.fmin):
            fmin_template = int(14*(90/mtot_msun))   # should also decrease this due to lmax!
            print "  SEOB starting frequencies need to be reduced for this event; trying ", fmin_template


is_analysis_precessing =False
if opts.approx == "SEOBNRv3" or opts.approx == "NRSur7dq2" or opts.approx == "NRSur7dq4" or (opts.approx == 'SEOBNv3_opt') or (opts.approx == 'IMRPhenomPv2') or (opts.approx =="SEOBNRv4P" ) or (opts.approx == "SEOBNRv4PHM") or ('SpinTaylor' in opts.approx):
        is_analysis_precessing=True

dirname_run = gwid+ "_" + opts.calibration+ "_"+ opts.approx+"_fmin" + str(fmin) +"_fmin-template"+str(fmin_template) +"_lmax"+str(opts.l_max) + "_"+opts.spin_magnitude_prior
if opts.online:
    dirname_run += "_onlineLLframes"
elif opts.use_online_psd:
    dirname_run += "_onlinePSD"
elif opts.link_bw_psds:
    dirname_run += "_fiducialBWpsd"
elif opts.make_bw_psds:
    dirname_run += "_manualBWpsd"
if opts.data_LI_seglen:
    dirname_run += "_LIseglen"+str(opts.data_LI_seglen)
if opts.assume_matter:
    dirname_run += "_with_matter"
if opts.no_matter:
    dirname_run += "_no_matter"
if opts.assume_highq:
    dirname_run+="_highq"
if opts.manual_postfix:
    dirname_run += opts.manual_postfix
if opts.playground_data:
    dirname_run = "playground_" + dirname_run
if not(opts.cip_sampler_method is None):
    dirname_run += "_" + opts.cip_sampler_method
if not(opts.cip_fit_method is None):
    dirname_run += "_" + opts.cip_fit_method
if opts.use_osg:
    dirname_run += '_OSG'
# Override run directory name
if opts.use_rundir:
    dirname_run = opts.use_rundir
os.mkdir(dirname_run)
os.chdir(dirname_run)


if not(opts.use_ini is None):
    if opts.use_coinc is None:
        print " coinc required for ini file operation at present "
        sys.exit(1)
    # Load in event dictionary
    event_dict = retrieve_event_from_coinc(opts.use_coinc)
    # Create relevant sim_xml file to hold parameters (does not parse coinc)
    P=lalsimutils.ChooseWaveformParams()
    P.m1 = event_dict["m1"]*lal.MSUN_SI; P.m2=event_dict["m2"]*lal.MSUN_SI; P.s1z = event_dict["s1z"]; P.s2z = event_dict["s2z"]
    # Load in ini file to select relevant fmin, fref [latter usually unused]
    config = ConfigParser.ConfigParser()
    config.read(opts.use_ini)
    fmin_vals ={}
    fmin_fiducial = -1
    ifo_list = eval(config.get('analysis','ifos'))
    for ifo in ifo_list:
        fmin_vals[ifo] = unsafe_config_get(config,['lalinference','flow'])[ifo]
        fmin_fiducial = fmin_vals[ifo]
    event_dict["IFOs"] = ifo_list
    print "IFO list from ini ", ifo_list
    P.fmin = fmin_fiducial
    P.fref = unsafe_config_get(config,['engine','fref'])
    # Write 'target_params.xml.gz' file
    lalsimutils.ChooseWaveformParams_array_to_xml([P], "target_params")



helper_psd_args = ''
srate=4096  # default, built into helper, unwise to go lower, LI will almost never do higher
if opts.make_bw_psds:
    helper_psd_args += " --assume-fiducial-psd-files --fmax " + str(srate/2-1)


# Create provenance info : we want run to be reproducible
if True:
        os.mkdir("reproducibility")
        # Write this script and its arguments
        import shutil, json
#        thisfile = os.path.realpath(__file__)
#        shutil.copyfile(thisfile, "reproducibility/the_script_used.py")
        argparse_dict = vars(opts)
        with open("reproducibility/the_arguments_used.json",'w') as f:
                json.dump(argparse_dict,f)
        # Write commits
        cmd = "(cd ${ILE_CODE_PATH}; git rev-parse HEAD) > reproducibility/RIFT.commit"
        os.system(cmd)
        module_list = ['gwsurrogate',  'NRSur7dq2', 'scipy', 'numpy', 'sklearn']
        with open("reproducibility/module_versions", 'w') as f:
                for name in module_list:
                    try:
                        exec("import "+ name+"; val = "+name+".__version__")
                        f.write(name +" " +val+"\n")
                    except:
                        print " No provenance for ", name

# Run helper command
npts_it = 500
cmd = " helper_LDG_Events.py --force-notune-initial-grid   --propose-fit-strategy --propose-ile-convergence-options --propose-initial-grid --fmin " + str(fmin) + " --fmin-template " + str(fmin_template) + " --working-directory " + base_dir + "/" + dirname_run  + helper_psd_args  + " --no-enforce-duration-bound "
if not(opts.gracedb_id is None) and (opts.use_ini is None):
    cmd +=" --use-legacy-gracedb --gracedb-id " + gwid 
elif  not(opts.event_time is None):
    cmd += " --event-time " + str(opts.event_time)
if opts.online:
        cmd += " --online "
if opts.playground_data:
        cmd += " --playground-data "
if opts.use_online_psd:
        cmd += " --use-online-psd "
if opts.data_LI_seglen:
        cmd += " --data-LI-seglen "+str(opts.data_LI_seglen)
if opts.assume_matter:
        cmd += " --assume-matter "
        npts_it = 1000
if opts.assume_highq:
    cmd+= ' --assume-highq  --force-grid-stretch-mc-factor 2'  # the mc range, tuned to equal-mass binaries, is probably too narrow. Workaround until fixed in helper
    npts_it =1000
#if is_event_bns and not opts.no_matter:
#        cmd += " --assume-matter "
#        npts_it = 1000
if is_analysis_precessing:
        cmd += " --assume-precessing-spin "
        npts_it = 1500
if opts.use_osg:
    cmd += " --use-osg "
    cmd += " --use-cvmfs-frames "  # only run with CVMFS data, otherwise very very painful
if opts.use_ini:
    cmd += " --use-ini " + opts.use_ini
    cmd += " --sim-xml target_params.xml.gz --event 0 "
    cmd += " --event-time " + str(event_dict["tref"])
    #
else:
    cmd += " --calibration-version " + opts.calibration 
if opts.use_online_psd_file:
    # Get IFO list from ini file
    import ConfigParser
    config = ConfigParser.ConfigParser()
    config.read(opts.use_ini)
    ifo_list = eval(config.get('analysis','ifos'))
    # Create command line arguments for those IFOs, so helper can correctly pass then downward
    for ifo in ifo_list:
        cmd+= " --psd-file {}={}".format(ifo,opts.use_online_psd_file)

print cmd
os.system(cmd)
#sys.exit(0)

# Create distance maximum (since that is NOT always chosen by the helper, and makes BNS runs needlessly much more painful!)
observing_run = 'O3'
if (opts.use_ini is None):
 try:
  with open("event.log",'r') as f:
    lines = f.readlines()
    for line in lines:
        if 'ime:' in line:  # look for Event time, Event Time, etc
            tref = float(line.split(' ')[-1])
            observing_run = get_observing_run(tref)
        if 'hirp' in line:
            mc_Msun = float(line.split(' ')[-1])
 except:
   print " Failure parsing event.log"
else:
    # use sim_xml produced above to generate necessary parameters
    t_ref = P.tref
    mc_Msun = P.extract_param('mc')/lal.MSUN_SI
snr_fac=1
#mc_Msun = P.extract_param('mc')/lal.MSUN_SI
try:
    dmax_guess =(1./snr_fac)* 2.5*2.26*typical_bns_range_Mpc[observing_run]* (mc_Msun/1.2)**(5./6.)
    dmax_guess = np.min([dmax_guess,10000]) # place ceiling
except:
    print " ===> Defaulting to maximum distance <=== "
    dmax_guess = 10000
# Last stage of commands done by other tools: too annoying to copy stuff over and run the next generation of the pipeline
instructions_ile = np.loadtxt("helper_ile_args.txt", dtype=str)  # should be one line
line = ' '.join(instructions_ile)
line += " --l-max " + str(opts.l_max) 
line += " --d-max " + str(dmax_guess)
if not 'NR' in opts.approx:
        line += " --approx " + opts.approx
elif opts.use_gwsurrogate and 'NRHybSur' in opts.approx:
        line += " --rom-group my_surrogates/nr_surrogates/ --rom-param NRHybSur3dq8.h5  "
elif opts.use_gwsurrogate and "NRSur7d" in opts.approx:
        line += " --rom-group my_surrogates/nr_surrogates/ --rom-param NRSur7dq2.h5  "
elif ("SEOBNR" in opts.approx) or ("NRHybSur" in opts.approx) or ("NRSur7d" in opts.approx): 
        line += " --approx " + opts.approx
else:
        print " Unknown approx ", opts.approx
        sys.exit(1)
if not(opts.manual_extra_ile_args is None):
    line += opts.manual_extra_ile_args
with open('args_ile.txt','w') as f:
        f.write(line)
os.system("cp helper_test_args.txt args_test.txt")

# CIP
#   - modify priors to be consistent with the spin priors used in the paper
#   - for the BNS, set chi_max
with open("helper_cip_arg_list.txt",'r') as f:
        raw_lines = f.readlines()


# Add arguments to the file we will use
instructions_cip = map(lambda x: x.rstrip().split(' '), raw_lines)#np.loadtxt("helper_cip_arg_list.txt", dtype=str)
n_iterations =0
lines  = []
for indx in np.arange(len(instructions_cip)):
    n_iterations += int(instructions_cip[indx][0])
    line = ' ' .join(instructions_cip[indx])
    line +=" --n-output-samples 10000 --n-eff 10000 --n-max 10000000   --downselect-parameter m2 --downselect-parameter-range [1,1000] "
    if not(opts.cip_fit_method is None):
        line = line.replace('--fit-method gp', '--fit-method ' + opts.cip_fit_method)
    if not (opts.cip_sampler_method is None):
        line += " --sampler-method "+opts.cip_sampler_method
    line += prior_args_lookup[opts.spin_magnitude_prior]
    if opts.hierarchical_merger_prior_1g:
        # Must use mtotal, q coordinates!  Change defaults
        line = line.replace('parameter mc', 'parameter mtot')
        line = line.replace('parameter delta_mc', 'parameter q')
        line += " --prior-tapered-mass-ratio "
    elif opts.hierarchical_merger_prior_2g:
        # Must use mtotal, q coordinates! Change defaults
        line = line.replace('parameter mc', 'parameter mtot')
        line = line.replace('parameter delta_mc', 'parameter q')
        line += " --prior-gaussian-mass-ratio --prior-gaussian-spin1-magnitude "   # should require precessing analysis
    elif opts.assume_highq:
        line += " --sampler-method GMM --internal-correlate-parameters 'mc,delta_mc,s1z' "
    if opts.approx in lalsimutils.waveform_approx_limit_dict:
        chi_max = lalsimutils.waveform_approx_limit_dict[opts.approx]["chi-max"]
        if not(opts.force_chi_max is None):
            chi_max = opts.force_chi_max
        q_min = lalsimutils.waveform_approx_limit_dict[opts.approx]["q-min"]
        eta_min = q_min/(1+q_min)**2
        line += " --chi-max {}  "
        # Parse arguments, impose limit based on the approximant used, as described above
#        import StringIO
        my_parser = argparse.ArgumentParser()
        my_parser.add_argument("--eta-range")
        my_opts, unknown_opts =my_parser.parse_known_args(line.split())
        eta_range_orig = eval(my_opts.eta_range)
        eta_range_revised = [np.max([eta_min,eta_range_orig[0]]),np.min([1,eta_range_orig[1]])]
        line=line.replace("--eta-range "+my_opts.eta_range,"--eta-range "+str(eta_range_revised))
        # Ideally, load in initial grid, and remove points outside the targeted range
        # IMPLEMENT THIS
    if opts.fit_save_gp:
        line += " --fit-save-gp my_gp "  # fiducial filename, stored in each iteration
    line += "\n"
    lines.append(line)

if opts.cip_quadratic_first:
    lines[0]=lines[0].replace(' --fit-method gp ', ' --fit-method quadratic ')
    lines[0]=lines[0].replace(' --parameter delta_mc ', ' --parameter eta ')   # almost without fail we are using mc, delta_mc, xi  as zeroth layer

with open("args_cip_list.txt",'w') as f: 
   for line in lines:
           f.write(line)

# Write test file
with open("args_test.txt",'w') as f:
        f.write("X --always-succeed --method lame  --parameter m1")


# Write puff file
puff_params = " --parameter mc --parameter delta_mc --parameter chieff_aligned "
puff_max_it =4
#  Read puff args from file, if present
try:
    with open("helper_puff_max_it.txt",'r') as f:
        puff_max_it = int(f.readline())
except:
    print " No puff file "
if opts.assume_matter:
    puff_params += " --parameter LambdaTilde "
    puff_max_it +=5   # make sure we resolve the correlations
if opts.assume_highq:
    puff_params = puff_params.replace(' delta_mc ', ' eta ')  # use natural coordinates in the high q strategy. May want to do this always
    puff_max_it +=3
with open("args_puff.txt",'w') as f:
        puff_args = puff_params + "--downselect-parameter chi1 --downselect-parameter-range [0,1] --downselect-parameter chi2 --downselect-parameter-range [0,1] "
        if opts.data_LI_seglen:
                puff_args+= " --enforce-duration-bound " +str(opts.data_LI_seglen)
        f.write("X " + puff_args)


# Overwrite grid if needed
if not (opts.manual_initial_grid is None):
    shutil.copyfile(opts.manual_initial_grid, "proposed-grid.xml.gz")

# Build DAG
cip_mem  = 30000
if opts.fit_method == 'rf':  # much lower memory requirement
    cip_mem = 4000
cmd ="create_event_parameter_pipeline_BasicIteration --request-gpu-ILE --ile-n-events-to-analyze 20 --input-grid proposed-grid.xml.gz --ile-exe  `which integrate_likelihood_extrinsic_batchmode`   --ile-args args_ile.txt --cip-args-list args_cip_list.txt --test-args args_test.txt --request-memory-CIP {} --request-memory-ILE 4096 --n-samples-per-job ".format(cip_mem) + str(npts_it) + " --working-directory `pwd` --n-iterations " + str(n_iterations) + " --n-copies 1" + " --puff-exe `which util_ParameterPuffball.py` --puff-cadence 1 --puff-max-it " + str(puff_max_it)+ " --puff-args args_puff.txt  --ile-retries "+ str(opts.ile_retries)
if opts.add_extrinsic:
    cmd += " --last-iteration-extrinsic --last-iteration-extrinsic-nsamples 8000 "
if opts.cip_explode_jobs:
   cmd+= " --cip-explode-jobs  " + str(opts.cip_explode_jobs)
if opts.use_osg:
    cmd += " --use-osg --use-singularity --use-cvmfs-frames --cache-file local.cache "   # run on the OSG, make sure to get frames (rather than try to transfer them).  Note with CVMFS frames we need to provide the cache
    cmd+= " --transfer-file-list  "+base_dir+"/"+dirname_run+"/helper_transfer_files.txt"
print cmd
os.system(cmd)



#! /usr/bin/env python
#
# USAGE
#  python util_JoinXML.py ~/overlap-grid.xml.gz  ~/overlap-grid.xml.gz  --output-file merged_output  --verbose X --force-approximant SEOBNRv2
import RIFT.lalsimutils as lalsimutils
import lalsimulation as lalsim
import numpy as np
import lal

rosDebug = True

lalsimutils.rosDebugMessagesContainer[0] = False

import optparse
parser = optparse.OptionParser()
parser.add_option("--output-file", default="merged_output", help="Filename, note xml.gz is appened automatically")
parser.add_option("--compare-parameter", action='append')
parser.add_option("--force-approximant",default=None)  # because there is a weird bug that sometimes hits me with type conversion
parser.add_option("--verbose")
opts, args = parser.parse_args()

# Loop over arguments, joining
P_list =[]
for fname in args:
    print " File name : ", fname
    P_list_here = lalsimutils.xml_to_ChooseWaveformParams_array(fname)
    print " Read n= ", len(P_list_here)
    for P in P_list_here:
        P.tref = 0.0  # insure it is ok  -- there have been problems with malformed times. DANGEROUS if used for another purpose than mine!
    print " Revised length before cutting: N = ", len(P_list_here)
    if opts.force_approximant:
        approx = lalsim.GetApproximantFromString(opts.force_approximant)
        print " Forcing approx = ", approx,  " AKA ", lalsim.GetStringFromApproximant(approx), " from ", opts.force_approximant
        for P in P_list_here:
            P.approx = approx
    P_list = P_list +  P_list_here

if opts.verbose:
    for P in P_list:
        P.print_params()
#        print P.approx


# Loop, removing duplicates sequentially
P_list_out =[]
n = len(P_list)
indxList = np.arange(n-1)
print indxList
param_names = opts.compare_parameter
for indx in indxList:
    print " Testing ", indx, "out of ",  n
    do_we_add_it = True
    for indx2 in np.arange(indx+1,n):
        are_all_parameters_the_same = True
        for param in param_names:
            param_val = P_list[indx].extract_param(param)
            param_val2 = P_list[indx2].extract_param(param)
            if rosDebug:
                print param, param_val, param_val2
            if not(param_val  == param_val2):
                are_all_parameters_the_same = False
        # conclude test for this index
        if are_all_parameters_the_same:
            do_we_add_it = False
            if rosDebug:
                print " Match between ", indx, " and ",  indx2
    if do_we_add_it:
        P_list_out = P_list_out + [P_list[indx]]

# Write output
print " Saving ", len(P_list_out)
lalsimutils.ChooseWaveformParams_array_to_xml(P_list_out, fname=opts.output_file) # note xml.gz is appended
